#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
###########################################################################

"""This package provides the TaurusGui class"""

__all__ = ["DockWidgetPanel", "TaurusGui"]

__docformat__ = 'restructuredtext'


import os
import sys
import copy
import weakref
import inspect

from lxml import etree

import taurus
from taurus import tauruscustomsettings
from taurus.external.qt import Qt
from taurus.qt.qtcore.configuration import BaseConfigurableClass
from taurus.qt.qtcore.communication import SharedDataManager
from taurus.qt.qtgui.util import TaurusWidgetFactory
from taurus.qt.qtgui.base import TaurusBaseWidget, TaurusBaseComponent
from taurus.qt.qtgui.container import TaurusMainWindow
from taurus.qt.qtgui.taurusgui.utils import (ExternalApp, PanelDescription,
                                             ToolBarDescription,
                                             AppletDescription)
from taurus.qt.qtgui.taurusgui.appsettingswizard import ExternalAppEditor
from taurus.qt.qtgui.panel import QDoubleListDlg
from taurus.qt.qtgui.util.ui import UILoadable
from taurus.qt.qtgui.taurusgui.utils import ExternalAppAction


@UILoadable(with_ui='ui')
class AssociationDialog(Qt.QDialog):
    '''A dialog for viewing and editing the associations between instruments
    and panels'''

    def __init__(self, parent, flags=None):
        if flags is None:
            flags = Qt.Qt.Widget
        Qt.QDialog.__init__(self, parent, flags)
        self.loadUi()

        self.refresh()
        self.ui.instrumentCB.activated.connect(self.onInstrumentChanged)
        self.ui.buttonBox.clicked.connect(self.onDialogButtonClicked)
        self.ui.refreshBT.clicked.connect(self.refresh)

    def refresh(self):
        currentinstrument = self.ui.instrumentCB.currentText()
        mainwindow = self.parent()

        self.associations = mainwindow.getAllInstrumentAssociations()

        # fill the comboboxes
        self.ui.instrumentCB.clear()
        self.ui.panelCB.clear()
        self.ui.instrumentCB.addItems(sorted(self.associations.keys()))
        self.ui.panelCB.addItems(['__[None]__'] + mainwindow.getPanelNames())

        # restore the index
        idx = self.ui.instrumentCB.findText(currentinstrument)
        if idx == -1 and self.ui.instrumentCB.count() > 0:
            idx = 0
        self.ui.instrumentCB.setCurrentIndex(idx)
        self.onInstrumentChanged(self.ui.instrumentCB.currentText())

    def onInstrumentChanged(self, instrumentname):
        instrumentname = unicode(instrumentname)
        panelname = self.associations.get(instrumentname)
        if panelname is None:
            self.ui.panelCB.setCurrentIndex(0)
            return
        else:
            idx = self.ui.panelCB.findText(panelname)
            self.ui.panelCB.setCurrentIndex(idx)

    def onDialogButtonClicked(self, button):
        role = self.ui.buttonBox.buttonRole(button)
        if role in (Qt.QDialogButtonBox.AcceptRole, Qt.QDialogButtonBox.ApplyRole):
            if self.ui.panelCB.currentIndex() > 0:
                panelname = unicode(self.ui.panelCB.currentText())
            else:
                panelname = None
            instrumentname = unicode(self.ui.instrumentCB.currentText())
            self.associations[instrumentname] = panelname
            self.parent().setInstrumentAssociation(instrumentname, panelname)


class DockWidgetPanel(Qt.QDockWidget, TaurusBaseWidget):
    '''
    This is an extended QDockWidget which provides some methods for being used
    as a "panel" of a TaurusGui application. Widgets of TaurusGui are inserted
    in the application by adding them to a DockWidgetPanel.
    '''

    def __init__(self, parent, widget, name, mainwindow):
        Qt.QDockWidget.__init__(self, None)
        TaurusBaseWidget.__init__(self, name, parent=parent)

        self.setAllowedAreas(Qt.Qt.TopDockWidgetArea)

        self.setWidget(widget)
        # self._widget = self.widget()  #keep a pointer that may change if the
        # widget changes
        name = unicode(name)
        self.setWindowTitle(name)
        self.setObjectName(name)
        self._custom = False

        # store a weakref of the main window
        self._mainwindow = weakref.proxy(mainwindow)

    def isCustom(self):
        return self._custom

    def setCustom(self, custom):
        self._custom = custom

    def isPermanent(self):
        return self._permanent

    def setPermanent(self, permanent):
        self._permanent = permanent

    def setWidgetFromClassName(self, classname, modulename=None):
        if self.getWidgetClassName() != classname:
            try:
                klass = TaurusWidgetFactory().getWidgetClass(classname)
                w = klass()
            except:
                try:
                    if classname is not None and '.' in classname:
                        mn, classname = classname.rsplit('.', 1)
                        modulename = ("%s.%s" %
                                      (modulename or '', mn)).strip('. ')
                    module = __import__(modulename, fromlist=[''])
                    klass = getattr(module, classname)
                    w = klass()
                except Exception, e:
                    raise RuntimeError(
                        'Cannot create widget from classname "%s". Reason: %s' % (classname, repr(e)))
            # set customwidgetmap if necessary
            if hasattr(w, 'setCustomWidgetMap'):
                w.setCustomWidgetMap(self._mainwindow.getCustomWidgetMap())
            self.setWidget(w)
            wname = "%s-%s" % (str(self.objectName()), str(classname))
            w.setObjectName(wname)

    def getWidgetModuleName(self):
        w = self.widget()
        if w is None:
            return ''
        return w.__module__

    def getWidgetClassName(self):
        w = self.widget()
        if w is None:
            return ''
        return w.__class__.__name__

    def applyConfig(self, configdict, depth=-1):
        # create the widget
        try:
            self.setWidgetFromClassName(configdict.get(
                'widgetClassName'), modulename=configdict.get('widgetModuleName', None))
            if isinstance(self.widget(), BaseConfigurableClass):
                self.widget().applyConfig(configdict['widget'])
        except Exception, e:
            self.info(
                'Failed to set the widget for this panel. Reason: %s' % repr(e))
            self.traceback(self.Debug)
            return
        TaurusBaseWidget.applyConfig(self, configdict, depth)

    def createConfig(self, *args, **kwargs):
        configdict = TaurusBaseWidget.createConfig(self, *args, **kwargs)
        configdict['widgetClassName'] = self.getWidgetClassName()
        configdict['widgetModuleName'] = self.getWidgetModuleName()
        if isinstance(self.widget(), BaseConfigurableClass):
            configdict['widget'] = self.widget().createConfig()
        return configdict

    def closeEvent(self, event):
        Qt.QDockWidget.closeEvent(self, event)
        TaurusBaseWidget.closeEvent(self, event)


class TaurusGui(TaurusMainWindow):
    '''
    This is main class for constructing the dynamic GUIs. TaurusGui is a
    specialised TaurusMainWindow which is able to handle "panels" and load
    configuration files.
    There are several ways of using TaurusGui. In the following we will give
    3 examples on how to create a simple GUI called "MyGui" which contains one
    panel called "Foo" and consisting of a `QWidget`:

    **Example 1: use declarative configuration files.**

    You can create a purely declarative configuration file to be interpreted by
    the standard `taurusgui` script::

        from taurus.qt.qtgui.taurusgui.utils import PanelDescription

        GUI_NAME = 'MyGui'
        panel = PanelDescription('Foo',
                                 classname='taurus.external.qt.Qt.QWidget')

    Note that this just a very simple example. For a much richer one, see the
    :mod:`taurus.qt.qtgui.taurusgui.conf.tgconf_example01`

    **Example 2: do everything programmatically.**

    A stand-alone python script that launches the gui when executed. No
    configuration file is used here. Panels and other components are added
    programatically::

        if __name__ == '__main__':
            from taurus.qt.qtgui.application import TaurusApplication
            from taurus.qt.qtgui.taurusgui import TaurusGui
            from taurus.external.qt import Qt
            app = TaurusApplication(app_name='MyGui')
            gui = TaurusGui()
            panel = Qt.QWidget()
            gui.createPanel(panel, 'Foo')
            gui.show()
            app.exec_()


    **Example 3: mixing declarative and programmatic ways**

    It is also possible to create a stand-alone python script which loads itself
    as a configuration file. In this way you can add things programmatically and
    at the same time use the declarative way::

        GUI_NAME = 'MyGui' # <-- declarative!
        if __name__ == '__main__':
            from taurus.qt.qtgui.application import TaurusApplication
            from taurus.qt.qtgui.taurusgui import TaurusGui
            from taurus.external.qt import Qt
            app = TaurusApplication()
            gui = TaurusGui(confname=__file__)
            panel = Qt.QWidget()
            gui.createPanel(panel, 'Foo')  # <-- programmatic!
            gui.show()
            app.exec_()

    '''

    SelectedInstrument = Qt.pyqtSignal(str)
    doorNameChanged = Qt.pyqtSignal(str)
    macroserverNameChanged = Qt.pyqtSignal(str)
    newShortMessage = Qt.pyqtSignal(str)

    IMPLICIT_ASSOCIATION = '__[IMPLICIT]__'

    def __init__(self, parent=None, confname=None, configRecursionDepth=None):
        TaurusMainWindow.__init__(self, parent, False, True)

        if configRecursionDepth is not None:
            self.defaultConfigRecursionDepth = configRecursionDepth

        self.__panels = {}
        self.__external_app = {}
        self.__external_app_actions = {}
        self._external_app_names = []
        self.__permanent_ext_apps = []
        self.__synoptics = []
        self.__instrumentToPanelMap = {}
        self.__panelToInstrumentMap = {}
        self.setDockNestingEnabled(True)

        self.registerConfigProperty(self._getPermanentExternalApps,
                                    self._setPermanentExternalApps,
                                    'permanentexternalapps')
        self.registerConfigProperty(
            self._getPermanentCustomPanels, self._setPermanentCustomPanels, 'permanentCustomPanels')
        self.registerConfigProperty(self.getAllInstrumentAssociations,
                                    self.setAllInstrumentAssociations, 'instrumentAssociation')

        from taurus import tauruscustomsettings
        self.setCustomWidgetMap(
            getattr(tauruscustomsettings, 'T_FORM_CUSTOM_WIDGET_MAP', {}))

        # Create a global SharedDataManager
        Qt.qApp.SDM = SharedDataManager(self)

        self.__initPanelsMenu()
        self.__initQuickAccessToolBar()
        self.__initJorgBar()
        self.__initSharedDataConnections()
        self.__initToolsMenu()
        self.__initViewMenu()
        self.__initPanelsToolBar()

        self.loadConfiguration(confname)

        # connect the main window itself as a reader/writer of "short messages"
        Qt.qApp.SDM.connectReader("shortMessage", self.onShortMessage)
        Qt.qApp.SDM.connectWriter("shortMessage", self, 'newShortMessage')

        # emit a short message informing that we are ready to go
        msg = '%s is ready' % Qt.qApp.applicationName()
        self.newShortMessage.emit(msg)

        if self.defaultConfigRecursionDepth >= 0:
            Qt.QMessageBox.information(self, "Fail-proof mode",
                                       ('Running in fail-proof mode.' +
                                        '\nLoading of potentially problematic settings is disabled.' +
                                        '\nSome panels may not be loaded or may ignore previous user configuration' +
                                        '\nThis will also apply when loading perspectives'),
                                       Qt.QMessageBox.Ok, Qt.QMessageBox.NoButton)

    def closeEvent(self, event):
        try:
            self.__macroBroker.removeTemporaryPanels()
        except:
            pass
        TaurusMainWindow.closeEvent(self, event)
        for n, panel in self.__panels.items():
            panel.closeEvent(event)
            panel.widget().closeEvent(event)
            if not event.isAccepted():
                result = Qt.QMessageBox.question(
                    self, 'Closing error',
                    "Panel '%s' cannot be closed. Proceed closing?" % n,
                    Qt.QMessageBox.Yes | Qt.QMessageBox.No)
                if result == Qt.QMessageBox.Yes:
                    event.accept()
                else:
                    break

    def __updatePanelsMenu(self):
        '''dynamically fill the panels menus'''
        panelsmenu = self.sender()
        permanent = (panelsmenu == self.__permPanelsMenu)
        panelsmenu.clear()
        panelnames = sorted(
            [n for n, p in self.__panels.items() if (p.isPermanent() == permanent)])
        for name in panelnames:
            panelsmenu.addAction(self.__panels[name].toggleViewAction())

    def __initPanelsMenu(self):
        # Panels menu
        self.__panelsMenu = Qt.QMenu('Panels', self)
        self.menuBar().insertMenu(self.helpMenu.menuAction(), self.__panelsMenu)
        self.hideAllPanelsAction = self.__panelsMenu.addAction(
            Qt.QIcon('actions:hide.svg'), "Hide all panels", self.hideAllPanels)
        self.showAllPanelsAction = self.__panelsMenu.addAction(
            Qt.QIcon('actions:show.svg'), "Show all panels", self.showAllPanels)
        self.newPanelAction = self.__panelsMenu.addAction(
            Qt.QIcon.fromTheme("window-new"), "New Panel...",
            self.createCustomPanel)
        self.removePanelAction = self.__panelsMenu.addAction(
            Qt.QIcon.fromTheme("edit-clear"), "Remove Panel...",
            self.removePanel)
        self.__panelsMenu.addAction(Qt.QIcon.fromTheme(
            "preferences-desktop-personal"),
            "Switch temporary/permanent status...",
            self.updatePermanentCustomPanels)
        # temporary and permanent panels submenus
        self.__panelsMenu.addSeparator()
        self.__permPanelsMenu = Qt.QMenu('Permanent Panels', self)
        self.__panelsMenu.addMenu(self.__permPanelsMenu)
        self.__permPanelsMenu.aboutToShow.connect(self.__updatePanelsMenu)
        self.__tempPanelsMenu = Qt.QMenu('Temporary Panels', self)
        self.__panelsMenu.addMenu(self.__tempPanelsMenu)
        self.__tempPanelsMenu.aboutToShow.connect(self.__updatePanelsMenu)
        self.__panelsMenu.addSeparator()

    def __initViewMenu(self):
        # the superclass may already have added stuff to the viewMenu
        self.viewMenu.addSeparator()
        # view locking
        self.viewMenu.addSeparator()
        self._lockviewAction = Qt.QAction(Qt.QIcon.fromTheme(
            "system-lock-screen"), "Lock View", self)
        self._lockviewAction.setCheckable(True)
        self._lockviewAction.toggled.connect(self.setLockView)
        self._lockviewAction.setChecked(not self.isModifiableByUser())
        self.viewMenu.addAction(self._lockviewAction)

    def __initPanelsToolBar(self):
        # Panels toolbar
        self.panelsToolBar = self.addToolBar("Panels")
        self.panelsToolBar.setObjectName("PanelsToolbar")
        self.panelsToolBar.addAction(self.newPanelAction)
        self.viewToolBarsMenu.addAction(self.panelsToolBar.toggleViewAction())

    def __initQuickAccessToolBar(self):
        self.quickAccessToolBar = self.addToolBar("Quick Access")
        self.quickAccessToolBar.setObjectName("quickAccessToolbar")
        self.quickAccessToolBar.setToolButtonStyle(
            Qt.Qt.ToolButtonTextBesideIcon)
        self.viewToolBarsMenu.addAction(
            self.quickAccessToolBar.toggleViewAction())

    def __initJorgBar(self):
        # Fancy Stuff ToolBar (aka Jorg's Bar ;) )
        self.jorgsBar = Qt.QToolBar('Fancy ToolBar')
        self.jorgsBar.setObjectName('jorgsToolBar')
        self.addToolBar(Qt.Qt.RightToolBarArea, self.jorgsBar)
        self.jorgsBar.setIconSize(Qt.QSize(60, 60))
        self.jorgsBar.setMovable(False)

    def __initSharedDataConnections(self):
        # register the TAURUSGUI itself as a writer/reader for several shared
        # data items
        splashScreen = self.splashScreen()
        if splashScreen is not None:
            self.splashScreen().showMessage("setting up shared data connections")
        Qt.qApp.SDM.connectWriter(
            "macroserverName", self, 'macroserverNameChanged')
        Qt.qApp.SDM.connectWriter("doorName", self, 'doorNameChanged')
        Qt.qApp.SDM.connectReader(
            "SelectedInstrument", self.onSelectedInstrument)
        Qt.qApp.SDM.connectWriter(
            "SelectedInstrument", self, 'SelectedInstrument')
        Qt.qApp.SDM.connectReader("executionStarted", self.setFocusToPanel)
        Qt.qApp.SDM.connectReader("selectedPerspective", self.loadPerspective)
        Qt.qApp.SDM.connectWriter(
            "perspectiveChanged", self, 'perspectiveChanged')

    def __initToolsMenu(self):
        if self.toolsMenu is None:
            self.toolsMenu = Qt.QMenu("Tools")
        tm = self.toolsMenu
        tm.addAction(Qt.QIcon("apps:preferences-system-session.svg"),
                     "manage instrument-panel associations", self.onShowAssociationDialog)
        tm.addAction(Qt.QIcon.fromTheme("document-save"),
                     "Export current Panel configuration to XML",
                     self.onExportCurrentPanelConfiguration)
        tm.addAction(Qt.QIcon("actions:data-transfer.svg"),
                     "Show Shared Data Manager connections", self.showSDMInfo)

        # tools->external apps submenu
        self.addExternalApplicationAction = self.externalAppsMenu.addAction(
            Qt.QIcon.fromTheme('list-add'),
            'Add external application launcher...',
            self.createExternalApp)
        self.removeExternalApplicationAction = self.externalAppsMenu.addAction(
            Qt.QIcon.fromTheme('list-remove'),
            'Remove external appication launcher...',
            self.removeExternalApp)
        self.externalAppsMenu.addSeparator()

    def createExternalApp(self):
        '''Add a new external application on execution time'''
        app_editor = ExternalAppEditor(self)
        name, xml, ok = app_editor.getDialog()
        if name in self._external_app_names:
            msg = ('The "%s" external application exists in your GUI.'
                   ' If you want to create a new one, '
                   'please use other text label' % name)
            taurus.warning(msg)
            return

        if ok:
            extapp = ExternalApp.fromXml(xml)
            action = extapp.getAction()
            action_name = str(action.text())
            self.__external_app[action_name] = extapp
            self._addExternalAppLauncher(name, action)

    def _addExternalAppLauncher(self, name, action):
        action_name = str(action.text())
        self.__external_app_actions[action_name] = action
        self.addExternalAppLauncher(action)
        self._external_app_names.append(name)

    def removeExternalApp(self, name=None):
        '''Remove the given external application from the GUI.

        :param name: (str or None) the name of the external application to be
                     removed
                     If None given, the user will be prompted
        '''
        apps = self.__external_app.keys() + self.__permanent_ext_apps
        if name is None:
            items = sorted(apps)
            msg1 = "Remove External application"
            msg2 = ("External application to be removed "
                    "(only custom external applications can be removed).")
            name, ok = Qt.QInputDialog.getItem(self, msg1, msg2, items, 0,
                                               False)
            if not ok:
                return
        name = unicode(name)
        if name not in apps:
            msg = ('Cannot remove the external application "%s"'
                   ' (not found)' % name)
            self.debug(msg)
            return
        if name in self.__external_app.keys():
            self.__external_app.pop(name)
        else:
            self.__permanent_ext_apps.remove(name)
        action = self.__external_app_actions.pop(name)
        self._external_app_names.remove(name)
        self.deleteExternalAppLauncher(action)
        self.debug('External application "%s" removed' % name)

    def setCustomWidgetMap(self, map):
        '''
        Sets the widget map that is used application-wide. This widget map will
        be used by default in all TaurusForm Panels belonging to this gui.

        :param map: (dict<str,Qt.QWidget>) a dictionary whose keys are device
                    type strings (e.g. see :class:`PyTango.DeviceInfo`) and
                    whose values are widgets to be used

        .. seealso:: :meth:`TaurusForm.setCustomWidgetMap`, :meth:`getCustomWidgetMap`
        '''
        self._customWidgetMap = map

    def getCustomWidgetMap(self):
        '''
        Returns the default map used to create custom widgets by the TaurusForms
        belonging to this GUI

        :return: (dict<str,Qt.QWidget>) a dictionary whose keys are device
                 type strings (i.e. see :class:`PyTango.DeviceInfo`) and whose
                 values are widgets to be used

        .. seealso:: :meth:`setCustomWidgetMap`
        '''
        return self._customWidgetMap

    def createConfig(self, *args, **kwargs):
        '''reimplemented from TaurusMainWindow.createConfig'''
        self.updatePermanentCustomPanels(showAlways=False)
        self.updatePermanentExternalApplications(showAlways=False)
        cfg = TaurusMainWindow.createConfig(self, *args, **kwargs)
        return cfg

    def removePanel(self, name=None):
        ''' remove the given panel from the GUI.

        .. note:: The panel; is actually removed from the current perspective.
                  If the panel is saved in other perspectives, it should be removed from
                  them as well.

        :param name: (str or None) the name of the panel to be removed
                     If None given, the user will be prompted
        '''
        if name is None:
            items = sorted(
                [n for n, p in self.__panels.iteritems() if p.isCustom()])
            name, ok = Qt.QInputDialog.getItem(self, "Remove Panel",
                                               "Panel to be removed (only custom panels can be removed).\n Important: you may want to save the perspective afterwards,\n and maybe remove the panel from other perspectives as well", items, 0, False)
            if not ok:
                return
        name = unicode(name)
        if name not in self.__panels:
            self.debug('Cannot remove panel "%s" (not found)' % name)
            return
        panel = self.__panels.pop(name)
        try:
            # in case the widget is a Taurus one and does some cleaning when
            # setting model to None
            panel.widget().setModel(None)
        except:
            pass

        self.unregisterConfigurableItem(name, raiseOnError=False)
        self.removeDockWidget(panel)
        panel.setParent(None)
        panel.setAttribute(Qt.Qt.WA_DeleteOnClose)
        panel.close()
        self.debug('Panel "%s" removed' % name)

    def createPanel(self, widget, name, floating=False, registerconfig=True, custom=False,
                    permanent=False, icon=None, instrumentkey=None):
        '''
        Creates a panel containing the given widget.

        :param wiget: (QWidget) the widget to be contained in the panel
        :param name: (str) the name of the panel. It will be used in tabs as well as for configuration
        :param floating: (bool) whether the panel should be docked or floating. (see note below)
        :param registerconfig: (bool) if True, the panel will be registered as a delegate for configuration
        :param custom: (bool) if True the panel is to be considered a "custom panel"
        :param permanent: (bool) set this to True for panels that need to be recreated when restoring the app
        :param icon: (QIcon) icon for the panel
        :param instrumentkey: (str) name of an instrument to which this panel is to be associated

        :return: (DockWidgetPanel) the created panel

        .. note:: On a previous version, there was a mandatory parameter called
                  `area` (which accepted a Qt.DockWidgetArea or None as values)
                  this parameter has now been substituted by the keyword
                  argument `floating`. In order to provide backwards
                  compatibility, the "floating" keyword argument stays at the
                  same position as the old `area` argument and if a Qt.DockWidgetArea
                  value is given, it will be interpreted as floating=True (while if
                  `None` is passed, it will be interpreted as floating=False.
        '''

        # backwards compatibility:
        if not isinstance(floating, bool):
            self.info(
                'Deprecation warning: please note that the "area" argument is deprecated. See TaurusGui.createPanel doc')
            floating = not(floating)

        name = unicode(name)
        if name in self.__panels:
            self.info('Panel with name "%s" already exists. Reusing.' % name)
            return self.__panels[name]

        # create a panel
        panel = DockWidgetPanel(None, widget, name, self)
        # we will only place panels in this area
        self.addDockWidget(Qt.Qt.TopDockWidgetArea, panel)
        if len(self.__panels) != 0:
            self.tabifyDockWidget(self.__panels.values()[-1], panel)

        panel.setFloating(floating)

        # associate this panel with an instrument
        if instrumentkey is not None:
            if instrumentkey == self.IMPLICIT_ASSOCIATION:
                # see if there is an item whose name is the same as that of the
                # panel
                for syn in self.__synoptics:
                    if name in syn.get_item_list():
                        self.setInstrumentAssociation(name, name)
                        break
            else:
                self.setInstrumentAssociation(instrumentkey, name)

        if icon is not None:
            panel.toggleViewAction().setIcon(icon)

        # set flags
        panel.setCustom(custom)
        panel.setPermanent(permanent)

        # register the panel for configuration
        if registerconfig:
            self.registerConfigDelegate(panel, name=name)
        self.__panels[name] = panel

        # connect the panel visibility changes
        panel.visibilityChanged.connect(self._onPanelVisibilityChanged)

        return panel

    def getPanel(self, name):
        '''get a panel object by name

        :return: (DockWidgetPanel)
        '''
        return self.__panels[unicode(name)]

    def getPanelNames(self):
        '''returns the names of existing panels

        :return: (list<str>)
        '''
        return copy.deepcopy(self.__panels.keys())

    def _setPermanentExternalApps(self, permExternalApps):
        '''creates empty panels for restoring custom panels.

        :param permCustomPanels: (list<str>) list of names of custom panels
        '''
        # first create the panels if they don't actually exist
        for name in permExternalApps:
            if name not in self._external_app_names:
                # create empty action
                self.__permanent_ext_apps.append(name)
                action = ExternalAppAction('', name)
                self._addExternalAppLauncher(name, action)

    def _getPermanentExternalApps(self):
        return self.__permanent_ext_apps

    def _setPermanentCustomPanels(self, permCustomPanels):
        '''creates empty panels for restoring custom panels.

        :param permCustomPanels: (list<str>) list of names of custom panels
        '''
        # first create the panels if they don't actually exist
        for name in permCustomPanels:
            if name not in self.__panels:
                self.createPanel(None, name, custom=True, permanent=True)

    def _getPermanentCustomPanels(self):
        '''
        returns a list of panel names for which the custom and permanent flags
        are True (i.e., those custom panels that should be stored in
        configuration and/or perspectives)

        :return: (list<str>)
        '''
        return [n for n, p in self.__panels.iteritems() if (p.isCustom() and p.isPermanent())]

    def updatePermanentCustomPanels(self, showAlways=True):
        '''
        Shows a dialog for selecting which custom panels should be permanently
        stored in the configuration.

        :param showAlways: (bool) forces showing the dialog even if there are no new custom Panels
        '''
        # check if there are some newly created panels that may be made
        # permanent
        perm = self._getPermanentCustomPanels()
        temp = [n for n, p in self.__panels.iteritems() if (
            p.isCustom() and not p.isPermanent())]
        if len(temp) > 0 or showAlways:
            dlg = QDoubleListDlg(winTitle='Stored panels',
                                 mainLabel='Select which of the panels should be stored',
                                 label1='Temporary (to be discarded)', label2='Permanent (to be stored)',
                                 list1=temp, list2=perm)
            result = dlg.exec_()
            if result == Qt.QDialog.Accepted:
                # update the permanent Custom Panels
                registered = self.getConfigurableItemNames()
                for name in dlg.getAll2():
                    if name not in registered:
                        self.__panels[name].setPermanent(True)
                        self.registerConfigDelegate(self.__panels[name], name)
                # unregister any panel that is temporary
                for name in dlg.getAll1():
                    self.__panels[name].setPermanent(False)
                    self.unregisterConfigurableItem(name, raiseOnError=False)

    def updatePermanentExternalApplications(self, showAlways=True):
        '''
        Shows a dialog for selecting which new externals applications
        should be permanently stored in the configuration.

        :param showAlways: (bool) forces showing the dialog
        '''
        # check if there are some newly created external applications that may
        # be made permanent
        #permanet_ext_app = list(self._external_app_names)
        if len(self.__external_app) > 0 or showAlways:
            msg = 'Select which of the external applications should be stored'
            dlg = QDoubleListDlg(winTitle='Stored external applications',
                                 mainLabel=msg,
                                 label1='Temporary (to be discarded)',
                                 label2='Permanent (to be stored)',
                                 list1=self.__external_app.keys(),
                                 list2=self.__permanent_ext_apps)
            result = dlg.exec_()
            if result == Qt.QDialog.Accepted:
                # update the temporally external applications
                for name in dlg.getAll2():
                    self.__permanent_ext_apps.append(str(name))
                    if name in self.__external_app:
                        self.__external_app.pop(str(name))

                for name in dlg.getAll1():
                    self.unregisterConfigurableItem("_extApp[%s]" % str(name),
                                                    raiseOnError=False)

    def createCustomPanel(self, paneldesc=None):
        '''
        Creates a panel from a Panel Description and sets it as "custom panel".

        :param paneldesc: (PanelDescription) description of the panel to be created

        .. seealso:: :meth:`createPanel`
        '''

        if paneldesc is None:
            from taurus.qt.qtgui.taurusgui import PanelDescriptionWizard
            paneldesc, ok = PanelDescriptionWizard.getDialog(
                self, extraWidgets=self._extraCatalogWidgets)
            if not ok:
                return
        w = paneldesc.getWidget(sdm=Qt.qApp.SDM, setModel=False)
        if hasattr(w, 'setCustomWidgetMap'):
            w.setCustomWidgetMap(self.getCustomWidgetMap())
        if paneldesc.model is not None:
            w.setModel(paneldesc.model)
        if isinstance(w, TaurusBaseComponent):
            w.setModifiableByUser(True)
            w.setModelInConfig(True)

        self.createPanel(w, paneldesc.name, floating=paneldesc.floating, custom=True,
                         registerconfig=False, instrumentkey=paneldesc.instrumentkey,
                         permanent=False)
        msg = 'Panel %s created. Drag items to it or use the context menu to customize it' % w.name
        self.newShortMessage.emit(msg)

    def createMainSynoptic(self, synopticname):
        '''
        Creates a synoptic panel and registers it as "SelectedInstrument"
        reader and writer (allowing  selecting instruments from synoptic
        '''
        try:
            jdwFileName = os.path.join(self._confDirectory, synopticname)
            from taurus.qt.qtgui.graphic import TaurusJDrawSynopticsView
            synoptic = TaurusJDrawSynopticsView()
            synoptic.setModel(jdwFileName)
            self.__synoptics.append(synoptic)
        except Exception, e:
            # print repr(e)
            msg = 'Error loading synoptic file "%s".\nSynoptic won\'t be available' % jdwFileName
            self.error(msg)
            self.traceback(level=taurus.Info)
            result = Qt.QMessageBox.critical(self, 'Initialization error', '%s\n\n%s' % (
                msg, repr(e)), Qt.QMessageBox.Abort | Qt.QMessageBox.Ignore)
            if result == Qt.QMessageBox.Abort:
                sys.exit()

        Qt.qApp.SDM.connectWriter(
            "SelectedInstrument", synoptic, "graphicItemSelected")
        Qt.qApp.SDM.connectReader(
            "SelectedInstrument", synoptic.selectGraphicItem)

        # find an unique (and short) name
        name = os.path.splitext(os.path.basename(synopticname))[0]
        if len(name) > 10:
            name = 'Syn'
        i = 2
        prefix = name
        while name in self.__panels:
            name = '%s_%i' % (prefix, i)
            i += 1

        synopticpanel = self.createPanel(synoptic, name, permanent=True,
                                         icon=Qt.QIcon.fromTheme(
                                             'image-x-generic'))
        toggleSynopticAction = synopticpanel.toggleViewAction()
        self.quickAccessToolBar.addAction(toggleSynopticAction)

    def createConsole(self, kernels):
        try:
            from taurus.qt.qtgui.console import TaurusConsole
        except ImportError:
            self.warning(
                'Cannot import taurus.qt.qtgui.console. The Console Panel will not be available')
            return
        console = TaurusConsole(kernels=kernels)
        consolePanel = self.createPanel(console, "Console", permanent=True,
                                        icon=Qt.QIcon.fromTheme(
                                            'utilities-terminal'))
        toggleConsoleAction = consolePanel.toggleViewAction()
        self.quickAccessToolBar.addAction(toggleConsoleAction)

    def createInstrumentsFromPool(self, macroservername):
        '''
        Creates a list of instrument panel descriptions by gathering the info
        from the Pool. Each panel is a TaurusForm grouping together all those
        elements that belong to the same instrument according to the Pool info

        :return: (list<PanelDescription>)
        '''
        instrument_dict = {}
        try:
            ms = taurus.Device(macroservername)
            instruments = ms.getElementsOfType('Instrument')
            if instruments is None:
                raise
        except Exception, e:
            msg = 'Could not fetch Instrument list from "%s"' % macroservername
            self.error(msg)
            result = Qt.QMessageBox.critical(self, 'Initialization error', '%s\n\n%s' % (
                msg, repr(e)), Qt.QMessageBox.Abort | Qt.QMessageBox.Ignore)
            if result == Qt.QMessageBox.Abort:
                sys.exit()
            return []
        for i in instruments.values():
            i_name = i.full_name
            #i_name, i_unknown, i_type, i_pools = i.split()
            i_view = PanelDescription(
                i_name, classname='TaurusForm', floating=False, model=[])
            instrument_dict[i_name] = i_view

        from operator import attrgetter
        pool_elements = sorted(ms.getElementsWithInterface(
            'Moveable').values(), key=attrgetter('name'))
        pool_elements += sorted(ms.getElementsWithInterface(
            'ExpChannel').values(), key=attrgetter('name'))
        pool_elements += sorted(ms.getElementsWithInterface(
            'IORegister').values(), key=attrgetter('name'))
        for elem in pool_elements:
            instrument = elem.instrument
            if instrument:
                i_name = instrument
                e_name = elem.full_name
                instrument_dict[i_name].model.append(e_name)
        # filter out empty panels
        ret = [instrument for instrument in instrument_dict.values()
               if len(instrument.model) > 0]
        return ret

    def __getVarFromXML(self, root, nodename, default=None):
        name = root.find(nodename)
        if name is None or name.text is None:
            return default
        else:
            return name.text

    def _importConfiguration(self, confname):
        '''returns the module corresponding to `confname` or to
        `tgconf_<confname>`. Note: the `conf` subdirectory of the directory in
        which taurusgui.py file is installed is temporally prepended to sys.path
        '''
        confsubdir = os.path.join(os.path.abspath(os.path.dirname(
            __file__)), 'conf')  # the path to a conf subdirectory of the place where taurusgui.py is
        oldpath = sys.path
        try:
            # add the conf subdirectory dir to the pythonpath
            sys.path = [confsubdir] + sys.path
            conf = __import__(confname)
        except ImportError:
            altconfname = "tgconf_%s" % confname
            try:
                conf = __import__(altconfname)
            except ImportError:
                msg = 'cannot import %s or %s' % (confname, altconfname)
                self.error(msg)
                Qt.QMessageBox.critical(
                    self, 'Initialization error', msg, Qt.QMessageBox.Abort)
                sys.exit()
        finally:
            sys.path = oldpath  # restore the previous sys.path
        return conf

    def loadConfiguration(self, confname):
        '''Reads a configuration file

        :param confname: (str or None) the  name of module located in the
                         PYTHONPATH or in the conf subdirectory of the directory
                         in which taurusgui.py file is installed.
                         This method will try to import <confname>.
                         If that fails, it will try to import
                         `tgconf_<confname>`.
                         Alternatively, `confname` can be the path to the
                         configuration module (not necessarily in the
                         PYTHONPATH).
                         `confname` can also be None, in which case a dummy
                         empty module will be used.
        '''

        # import the python config file
        try:
            if confname is None:
                import types
                conf = types.ModuleType(
                    '__dummy_conf_module__')  # dummy module
                confname = str(Qt.qApp.applicationName())
                self._confDirectory = ''
            elif os.path.exists(confname):  # if confname is a dir or file name
                import imp
                path, name = os.path.split(confname)
                name, _ = os.path.splitext(name)
                try:
                    f, filename, data = imp.find_module(name, [path])
                    conf = imp.load_module(name, f, filename, data)
                    confname = name
                except ImportError:
                    conf = self._importConfiguration(confname)
                self._confDirectory = os.path.dirname(conf.__file__)
            else:  # if confname is not a dir name, we assume it is a module name in the python path
                conf = self._importConfiguration(confname)
                self._confDirectory = os.path.dirname(conf.__file__)
        except Exception, e:
            import traceback
            msg = 'Error loading configuration: %s' % traceback.format_exc()  # repr(e)
            self.error(msg)
            Qt.QMessageBox.critical(
                self, 'Initialization error', msg, Qt.QMessageBox.Abort)
            sys.exit()

        # Get the xml root node from the xml configuration file
        XML_CONFIG = getattr(conf, 'XML_CONFIG', None)
        if XML_CONFIG is None:
            self._xmlConfigFileName = None
        else:
            self._xmlConfigFileName = os.path.join(
                self._confDirectory, XML_CONFIG)
        # default fallback (in case of I/O or parse errors)
        xmlroot = etree.fromstring('<root></root>')
        if XML_CONFIG is not None:
            try:
                # If a relative name was given, the conf directory will be used
                # as base path
                xmlfname = os.path.join(self._confDirectory, XML_CONFIG)
                xmlFile = open(xmlfname, 'r')
                xmlstring = xmlFile.read()
                xmlFile.close()
                xmlroot = etree.fromstring(xmlstring)
            except Exception, e:
                msg = 'Error reading the XML file: "%s"' % xmlfname
                self.error(msg)
                self.traceback(level=taurus.Info)
                result = Qt.QMessageBox.critical(self, 'Initialization error', '%s\nReason:"%s"' % (
                    msg, repr(e)), Qt.QMessageBox.Abort | Qt.QMessageBox.Ignore)
                if result == Qt.QMessageBox.Abort:
                    sys.exit()

        # General Qt application settings and jorgs bar logos
        APPNAME = getattr(conf, 'GUI_NAME', self.__getVarFromXML(
            xmlroot, "GUI_NAME", confname))
        ORGNAME = getattr(conf, 'ORGANIZATION', self.__getVarFromXML(
            xmlroot, "ORGANIZATION", str(Qt.qApp.organizationName()) or 'Taurus'))
        CUSTOMLOGO = getattr(conf, 'CUSTOM_LOGO', getattr(
            conf, 'LOGO', self.__getVarFromXML(xmlroot, "CUSTOM_LOGO", 'logos:taurus.png')))
        if Qt.QFile.exists(CUSTOMLOGO):
            customIcon = Qt.QIcon(CUSTOMLOGO)
        else:
            customIcon = Qt.QIcon(os.path.join(
                self._confDirectory, CUSTOMLOGO))
        Qt.qApp.setApplicationName(APPNAME)
        Qt.qApp.setOrganizationName(ORGNAME)
        Qt.QApplication.instance().basicConfig()

        logo = getattr(tauruscustomsettings, 'ORGANIZATION_LOGO',
                       "logos:taurus.png")
        ORGANIZATIONLOGO = getattr(conf, 'ORGANIZATION_LOGO',
                                   self.__getVarFromXML(xmlroot,
                                                        "ORGANIZATION_LOGO",
                                                        logo))
        ##
        if Qt.QFile.exists(ORGANIZATIONLOGO):
            organizationIcon = Qt.QIcon(ORGANIZATIONLOGO)
        else:
            organizationIcon = Qt.QIcon(os.path.join(
                self._confDirectory, ORGANIZATIONLOGO))

        # if required, enforce that only one instance of this GUI can be run
        SINGLEINSTANCE = getattr(conf, 'SINGLE_INSTANCE', (self.__getVarFromXML(
            xmlroot, "SINGLE_INSTANCE", 'True').lower() == 'true'))
        if SINGLEINSTANCE:
            if not self.checkSingleInstance():
                msg = 'Only one istance of %s is allowed to run the same time' % (
                    APPNAME)
                self.error(msg)
                Qt.QMessageBox.critical(
                    self, 'Multiple copies', msg, Qt.QMessageBox.Abort)
                sys.exit(1)

        # some initialization
        self.resetQSettings()
        self.setWindowTitle(APPNAME)
        self.setWindowIcon(customIcon)
        self.jorgsBar.addAction(organizationIcon, ORGNAME)
        self.jorgsBar.addAction(customIcon, APPNAME)

        # get custom widget catalog entries
        # @todo: support also loading from xml
        EXTRA_CATALOG_WIDGETS = getattr(conf, 'EXTRA_CATALOG_WIDGETS', [])
        self._extraCatalogWidgets = []
        for classname, pixmapname in EXTRA_CATALOG_WIDGETS:
            # If a relative file name is given, the conf directory will be used
            # as base path
            if pixmapname and not Qt.QFile.exists(pixmapname):
                pixmapname = os.path.join(self._confDirectory, pixmapname)
            self._extraCatalogWidgets.append((classname, pixmapname))

        # manual panel
        MANUAL_URI = getattr(conf, 'MANUAL_URI',
                             self.__getVarFromXML(xmlroot, "MANUAL_URI",
                                                  taurus.Release.url))
        self.setHelpManualURI(MANUAL_URI)

        self.createPanel(self.helpManualBrowser, 'Manual', permanent=True,
                         icon=Qt.QIcon.fromTheme('help-browser'))

        # configure the macro infrastructure
        MACROSERVER_NAME = getattr(conf, 'MACROSERVER_NAME', self.__getVarFromXML(
            xmlroot, "MACROSERVER_NAME", None))
        MACRO_PANELS = getattr(conf, 'MACRO_PANELS', self.__getVarFromXML(
            xmlroot, "MACRO_PANELS", True))
        # macro infrastructure will only be created if MACROSERVER_NAME is set
        if MACRO_PANELS and MACROSERVER_NAME is not None:
            from taurus.qt.qtgui.taurusgui import MacroBroker
            self.__macroBroker = MacroBroker(self)
        if MACROSERVER_NAME:
            self.macroserverNameChanged.emit(MACROSERVER_NAME)

        DOOR_NAME = getattr(conf, 'DOOR_NAME',
                            self.__getVarFromXML(xmlroot, "DOOR_NAME", ''))
        if DOOR_NAME:
            self.doorNameChanged.emit(DOOR_NAME)

        MACROEDITORS_PATH = getattr(conf, 'MACROEDITORS_PATH', self.__getVarFromXML(
            xmlroot, "MACROEDITORS_PATH", ''))
        if MACROEDITORS_PATH:
            from sardana.taurus.qt.qtgui.extra_macroexecutor.macroparameterseditor.macroparameterseditor import ParamEditorManager
            ParamEditorManager().parsePaths(MACROEDITORS_PATH)
            ParamEditorManager().browsePaths()

        # Synoptics
        SYNOPTIC = getattr(conf, 'SYNOPTIC', None)
        if isinstance(SYNOPTIC, basestring):  # old config file style
            self.warning(
                'Deprecated usage of SYNOPTIC keyword (now it expects a list of paths). Please update your configuration file to: "SYNOPTIC=[\'%s\']".' % SYNOPTIC)
            SYNOPTIC = [SYNOPTIC]
        if SYNOPTIC is None:  # we look in the xml config file if not present in the python config
            SYNOPTIC = []
            node = xmlroot.find("SYNOPTIC")
            if (node is not None) and (node.text is not None):
                for child in node:
                    s = child.get("str")
                    # we do not append empty strings
                    if s is not None and len(s):
                        SYNOPTIC.append(s)
        for s in SYNOPTIC:
            self.createMainSynoptic(s)

        # Get panel descriptions from pool if required
        INSTRUMENTS_FROM_POOL = getattr(conf, 'INSTRUMENTS_FROM_POOL', (self.__getVarFromXML(
            xmlroot, "INSTRUMENTS_FROM_POOL", 'False').lower() == 'true'))
        if INSTRUMENTS_FROM_POOL:
            try:
                self.splashScreen().showMessage("Gathering Instrument info from Pool")
            except AttributeError:
                pass
            POOLINSTRUMENTS = self.createInstrumentsFromPool(
                MACROSERVER_NAME)  # auto create instruments from pool
        else:
            POOLINSTRUMENTS = []

        CONSOLE = getattr(conf, 'CONSOLE', self.__getVarFromXML(
            xmlroot, "CONSOLE", ['ipython']))
        if CONSOLE:
            self.createConsole(CONSOLE)

        # get custom panel descriptions from the python config file
        CUSTOM_PANELS = [obj for name, obj in inspect.getmembers(
            conf) if isinstance(obj, PanelDescription)]

        # add custom panel descriptions from xml config
        panelDescriptions = xmlroot.find("PanelDescriptions")
        if (panelDescriptions is not None):
            for child in panelDescriptions:
                if (child.tag == "PanelDescription"):
                    pd = PanelDescription.fromXml(etree.tostring(child))
                    if pd is not None:
                        CUSTOM_PANELS.append(pd)

        # create panels based on the panel descriptions gathered before
        for p in CUSTOM_PANELS + POOLINSTRUMENTS:
            try:
                try:
                    self.splashScreen().showMessage("Creating panel %s" % p.name)
                except AttributeError:
                    pass
                w = p.getWidget(sdm=Qt.qApp.SDM, setModel=False)
                if hasattr(w, 'setCustomWidgetMap'):
                    w.setCustomWidgetMap(self.getCustomWidgetMap())
                if p.model is not None:
                    w.setModel(p.model)
                if p.instrumentkey is None:
                    instrumentkey = self.IMPLICIT_ASSOCIATION
                # the pool instruments may change when the pool config changes,
                # so we do not store their config
                registerconfig = p not in POOLINSTRUMENTS
                # create a panel
                self.createPanel(w, p.name, floating=p.floating, registerconfig=registerconfig,
                                 instrumentkey=instrumentkey, permanent=True)
            except Exception, e:
                msg = 'Cannot create panel %s' % getattr(
                    p, 'name', '__Unknown__')
                self.error(msg)
                self.traceback(level=taurus.Info)
                result = Qt.QMessageBox.critical(self, 'Initialization error', '%s\n\n%s' % (
                    msg, repr(e)), Qt.QMessageBox.Abort | Qt.QMessageBox.Ignore)
                if result == Qt.QMessageBox.Abort:
                    sys.exit()

        # get custom toolbars descriptions from the python config file
        CUSTOM_TOOLBARS = [obj for name, obj in inspect.getmembers(
            conf) if isinstance(obj, ToolBarDescription)]

        # add custom toolbar descriptions from xml config
        toolBarDescriptions = xmlroot.find("ToolBarDescriptions")
        if (toolBarDescriptions is not None):
            for child in toolBarDescriptions:
                if (child.tag == "ToolBarDescription"):
                    d = ToolBarDescription.fromXml(etree.tostring(child))
                    if d is not None:
                        CUSTOM_TOOLBARS.append(d)

        # create toolbars based on the descriptions gathered before
        for d in CUSTOM_TOOLBARS:
            try:
                try:
                    self.splashScreen().showMessage("Creating Toolbar %s" % d.name)
                except AttributeError:
                    pass
                w = d.getWidget(sdm=Qt.qApp.SDM, setModel=False)
                if d.model is not None:
                    w.setModel(d.model)
                w.setWindowTitle(d.name)
                # add the toolbar to the window
                self.addToolBar(w)
                # add the toggleview action to the view menu
                self.viewToolBarsMenu.addAction(w.toggleViewAction())
                # register the toolbar as delegate if it supports it
                if isinstance(w, BaseConfigurableClass):
                    self.registerConfigDelegate(w, d.name)

            except Exception, e:
                msg = 'Cannot add toolbar %s' % getattr(
                    d, 'name', '__Unknown__')
                self.error(msg)
                self.traceback(level=taurus.Info)
                result = Qt.QMessageBox.critical(self, 'Initialization error', '%s\n\n%s' % (
                    msg, repr(e)), Qt.QMessageBox.Abort | Qt.QMessageBox.Ignore)
                if result == Qt.QMessageBox.Abort:
                    sys.exit()

        CUSTOM_APPLETS = []
        # for backwards compatibility
        MONITOR = getattr(
            conf, 'MONITOR', self.__getVarFromXML(xmlroot, "MONITOR", []))
        if MONITOR:
            CUSTOM_APPLETS.append(AppletDescription(
                'monitor', classname='TaurusMonitorTiny', model=MONITOR))

        # get custom applet descriptions from the python config file
        CUSTOM_APPLETS += [obj for name, obj in inspect.getmembers(
            conf) if isinstance(obj, AppletDescription)]

        # add applet descriptions from xml config
        appletDescriptions = xmlroot.find("AppletDescriptions")
        if (appletDescriptions is not None):
            for child in appletDescriptions:
                if (child.tag == "AppletDescription"):
                    d = AppletDescription.fromXml(etree.tostring(child))
                    if d is not None:
                        CUSTOM_APPLETS.append(d)

        # create applet based on the descriptions gathered before
        for d in CUSTOM_APPLETS:
            try:
                try:
                    self.splashScreen().showMessage("Creating applet %s" % d.name)
                except AttributeError:
                    pass
                w = d.getWidget(sdm=Qt.qApp.SDM, setModel=False)
                if d.model is not None:
                    w.setModel(d.model)
                # add the widget to the applets toolbar
                self.jorgsBar.addWidget(w)
                # register the toolbar as delegate if it supports it
                if isinstance(w, BaseConfigurableClass):
                    self.registerConfigDelegate(w, d.name)
            except Exception, e:
                msg = 'Cannot add applet %s' % getattr(
                    d, 'name', '__Unknown__')
                self.error(msg)
                self.traceback(level=taurus.Info)
                result = Qt.QMessageBox.critical(self, 'Initialization error', '%s\n\n%s' % (
                    msg, repr(e)), Qt.QMessageBox.Abort | Qt.QMessageBox.Ignore)
                if result == Qt.QMessageBox.Abort:
                    sys.exit()

        # add external applications from both the python and the xml config
        # files
        EXTERNAL_APPS = [obj for name, obj in inspect.getmembers(
            conf) if isinstance(obj, ExternalApp)]

        externalAppsNode = xmlroot.find("ExternalApps")
        if (externalAppsNode is not None):
            for child in externalAppsNode:
                if (child.tag == "ExternalApp"):
                    ea = ExternalApp.fromXml(etree.tostring(child))
                    if ea is not None:
                        EXTERNAL_APPS.append(ea)

        for a in EXTERNAL_APPS:
            self._external_app_names.append(str(a.getAction().text()))
            self.addExternalAppLauncher(a.getAction())

        # get the "factory settings" filename. By default, it is called
        # "default.ini" and resides in the configuration dir
        INIFILE = getattr(conf, 'INIFILE', self.__getVarFromXML(
            xmlroot, "INIFILE", "default.ini"))
        # if a relative name is given, the conf dir is used as the root path
        iniFileName = os.path.join(self._confDirectory, INIFILE)

        # read the settings (or the factory settings if the regular file is not
        # found)
        msg = "Loading previous state"
        if self.defaultConfigRecursionDepth >= 0:
            msg += " in Fail Proof mode"
        try:
            self.splashScreen().showMessage(msg)
        except AttributeError:
            pass
        self.loadSettings(factorySettingsFileName=iniFileName)

    def setLockView(self, locked):
        self.setModifiableByUser(not locked)

    def setModifiableByUser(self, modifiable):
        if modifiable:
            dwfeat = Qt.QDockWidget.AllDockWidgetFeatures
        else:
            dwfeat = Qt.QDockWidget.NoDockWidgetFeatures
        for panel in self.__panels.values():
            panel.toggleViewAction().setEnabled(modifiable)
            panel.setFeatures(dwfeat)
        for action in (self.newPanelAction, self.showAllPanelsAction,
                       self.hideAllPanelsAction,
                       self.addExternalApplicationAction,
                       self.removeExternalApplicationAction,
                       ):
            action.setEnabled(modifiable)

        self._lockviewAction.setChecked(not modifiable)
        TaurusMainWindow.setModifiableByUser(self, modifiable)

    def onShortMessage(self, msg):
        ''' Slot to be called when there is a new short message. Currently, the only action
        taken when there is a new message is to display it in the main window status bar.

        :param msg: (str) the short descriptive message to be handled
        '''
        self.statusBar().showMessage(msg)

    def hideAllPanels(self):
        '''hides all current panels'''
        for panel in self.__panels.itervalues():
            panel.hide()

    def showAllPanels(self):
        '''shows all current panels'''
        for panel in self.__panels.itervalues():
            panel.show()

    def onShowAssociationDialog(self):
        '''launches the instrument-panel association dialog (modal)'''
        dlg = AssociationDialog(self)
        Qt.qApp.SDM.connectWriter(
            "SelectedInstrument", dlg.ui.instrumentCB, "activated(QString)")
        dlg.exec_()
        Qt.qApp.SDM.disconnectWriter(
            "SelectedInstrument", dlg.ui.instrumentCB, "activated(QString)")

    def getInstrumentAssociation(self, instrumentname):
        '''
        Returns the panel name associated to an instrument name

        :param instrumentname: (str or None) The name of the instrument whose associated panel is wanted

        :return: (str or None) the associated panel name (or None).
        '''
        return self.__instrumentToPanelMap.get(instrumentname, None)

    def setInstrumentAssociation(self, instrumentname, panelname):
        '''
        Sets the panel name associated to an instrument

        :param instrumentname: (str) The name of the instrument
        :param panelname: (str or None) The name of the associated
                          panel or None to remove the association
                          for this instrument.
        '''
        instrumentname = unicode(instrumentname)
        # remove a previous association if it exists
        oldpanelname = self.__instrumentToPanelMap.get(instrumentname, None)
        self.__panelToInstrumentMap.pop(oldpanelname, None)

        # create the new association
        self.__instrumentToPanelMap[instrumentname] = panelname
        if panelname is not None:
            self.__panelToInstrumentMap[panelname] = instrumentname

    def getAllInstrumentAssociations(self):
        '''
        Returns the dictionary of instrument-panel associations

        :return: (dict<str,str>) a dict whose keys are the instruments known to the gui
                 and whose values are the corresponding associated panels (or None).
        '''
        return copy.deepcopy(self.__instrumentToPanelMap)

    def setAllInstrumentAssociations(self, associationsdict, clearExisting=False):
        '''
        Sets the dictionary of instrument-panel associations.
        By default, it keeps any existing association not present in the associationsdict.

        :param associationsdict: (dict<str,str>) a dict whose keys are the instruments names
                                 and whose values are the corresponding associated panels (or None)
        :param clearExisting: (bool) if True, the the existing asociations are cleared.
                              If False (default) existing associations are
                              updated with those in associationsdict
        '''
        if clearExisting:
            self.__instrumentToPanelMap = copy.deepcopy(associationsdict)
        else:
            self.__instrumentToPanelMap.update(copy.deepcopy(associationsdict))
        self.__panelToInstrumentMap = {}
        for k, v in self.__instrumentToPanelMap.iteritems():
            self.__panelToInstrumentMap[v] = k

    def _onPanelVisibilityChanged(self, visible):
        if visible:
            panelname = unicode(self.sender().objectName())
            instrumentname = self.__panelToInstrumentMap.get(panelname)
            if instrumentname is not None:
                self.SelectedInstrument.emit(instrumentname)

    def onSelectedInstrument(self, instrumentname):
        ''' Slot to be called when the selected instrument has changed (e.g. by user
        clicking in the synoptic)

        :param instrumentname: (str) The name that identifies the instrument.
        '''
        instrumentname = unicode(instrumentname)
        panelname = self.getInstrumentAssociation(instrumentname)
        self.setFocusToPanel(panelname)

    def setFocusToPanel(self, panelname):
        ''' Method that sets a focus for panel passed via an argument

        :param panelname: (str) The name that identifies the panel.
                               This name must be unique within the panels in the GUI.
        '''
        panelname = unicode(panelname)
        try:
            panel = self.__panels[panelname]
            panel.show()
            panel.setFocus()
            panel.raise_()
        except KeyError:
            pass

    def tabifyArea(self, area):
        ''' tabifies all panels in a given area.

        :param area: (Qt.DockWidgetArea)

        .. warning:: This method is deprecated
        '''
        raise DeprecationWarning(
            'tabifyArea is no longer supported (now all panels reside in the same DockWidget Area)')
        panels = self.findPanelsInArea(area)
        if len(panels) < 2:
            return
        p0 = panels[0]
        for p in panels[1:]:
            self.tabifyDockWidget(p0, p)

    def findPanelsInArea(self, area):
        ''' returns all panels in the given area

        :param area: (QMdiArea, Qt.DockWidgetArea, 'FLOATING' or None). If
                     area=='FLOATING', the dockwidgets that are floating will be
                     returned.
        :param area:  (Qt.DockWidgetArea or str )

        .. warning:: This method is deprecated
        '''
        raise DeprecationWarning(
            'findPanelsInArea is no longer supported (now all panels reside in the same DockWidget Area)')
        if area == 'FLOATING':
            return [p for p in self.__panels.values() if p.isFloating()]
        else:
            return [p for p in self.__panels.values() if self.dockWidgetArea(p) == area]

    @classmethod
    def getQtDesignerPluginInfo(cls):
        '''TaurusGui is not to be in designer '''
        return None

    def onShowManual(self, anchor=None):
        '''reimplemented from :class:`TaurusMainWindow` to show the manual in a panel (not just a dockwidget)'''
        self.setFocusToPanel('Manual')

    def onExportCurrentPanelConfiguration(self, fname=None):

        if fname is None:
            fname = self._xmlConfigFileName

        if self._xmlConfigFileName is None:
            xmlroot = etree.Element("taurusgui_config")
        else:
            try:
                f = open(self._xmlConfigFileName, 'r')
                xmlroot = etree.fromstring(f.read())
                f.close()
            except Exception, e:
                self.error('Cannot parse file "%s": %s',
                           self._xmlConfigFileName, str(e))
                return

        # retrieve/create the PanelDescriptions node
        panelDescriptionsNode = xmlroot.find("PanelDescriptions")
        if panelDescriptionsNode is None:
            panelDescriptionsNode = etree.SubElement(
                xmlroot, "PanelDescriptions")

        # Get all custom panels
        dlg = QDoubleListDlg(winTitle='Export Panels to XML',
                             mainLabel='Select which of the custom panels you want to export as xml configuration',
                             label1='Not Exported', label2='Exported',
                             list1=[n for n, p in self.__panels.iteritems() if p.isCustom()], list2=[])
        result = dlg.exec_()
        if result != Qt.QDialog.Accepted:
            return
        exportlist = dlg.getAll2()

        # create xml for those to be exported
        registered = self.getConfigurableItemNames()
        for name in exportlist:
            panel = self.__panels[name]
            if name not in registered:
                panel.setPermanent(True)
                self.registerConfigDelegate(panel, name)
            panelxml = PanelDescription.fromPanel(panel).toXml()
            panelDescriptionsNode.append(etree.fromstring(panelxml))
        xml = etree.tostring(xmlroot, pretty_print=True)

        # write to file
        while True:
            if fname is None:
                fname = Qt.QFileDialog.getSaveFileName(
                    self, "Open File", fname or self._confDirectory, self.tr("XML files (*.xml)"))
                if not fname:
                    return
            fname = str(fname)
            # backup the file
            if os.path.exists(fname):
                import shutil
                try:
                    bckname = "%s.orig" % fname
                    shutil.copy(fname, bckname)
                except:
                    self.warning(
                        "%s will be overwritten but I could not create a backup in %s", fname, bckname)
            # write the data
            try:
                f = open(fname, 'w')
                f.write(xml)
                f.close()
                break
            except Exception, e:
                msg = 'Cannot write to %s: %s' % (fname, str(e))
                self.error(msg)
                Qt.QMessageBox.warning(
                    self, "I/O problem", msg + '\nChoose a different location.', Qt.QMessageBox.Ok, Qt.QMessageBox.NoButton)
                fname = None

        hint = "XML_CONFIG = '%s'" % os.path.relpath(
            fname, self._confDirectory)
        msg = 'Configuration written in %s' % fname
        self.info(msg)
        Qt.QMessageBox.information(self, "Configuration updated",
                                   msg + '\nMake sure that the .py configuration file in %s contains\n%s' % (
                                       self._confDirectory, hint),
                                   Qt.QMessageBox.Ok, Qt.QMessageBox.NoButton)

        return

    def showSDMInfo(self):
        '''pops up a dialog showing the current information from the Shared Data Manager'''
        #w = Qt.QMessageBox( self)
        text = 'Currently managing %i shared data objects:\n%s' % (
            len(Qt.qApp.SDM.activeDataUIDs()), ', '.join(Qt.qApp.SDM.activeDataUIDs()))
        nfo = Qt.qApp.SDM.info()
        w = Qt.QMessageBox(Qt.QMessageBox.Information, 'Shared Data Manager Information', text,
                           buttons=Qt.QMessageBox.Close, parent=self)
        w.setDetailedText(nfo)
        w.show()
        self.info(nfo)


#------------------------------------------------------------------------------
def main():
    import sys
    import taurus
    from taurus.core.util import argparse
    from taurus.qt.qtgui.application import TaurusApplication

    taurus.info('Starting execution of TaurusGui')

    parser = argparse.get_taurus_parser()
    parser.set_usage("%prog [options] confname")
    parser.set_description("The taurus GUI application")
    parser.add_option("", "--config-dir", dest="config_dir", default=None,
                      help="use the given configuration directory for initialization")
    parser.add_option("", "--new-gui", action="store_true", dest="new_gui", default=None,
                      help="launch a wizard for creating a new TaurusGUI application")
    parser.add_option("", "--fail-proof", action="store_true", dest="fail_proof", default=None,
                      help="launch in fail proof mode (it prevents potentially problematic configs from being loaded)")

    app = TaurusApplication(cmd_line_parser=parser, app_name="taurusgui",
                            app_version=taurus.Release.version)
    args = app.get_command_line_args()
    options = app.get_command_line_options()

    if options.new_gui:  # launch app settings wizard instead of taurusgui
        from taurus.qt.qtgui.taurusgui import AppSettingsWizard
        wizard = AppSettingsWizard()
        wizard.show()
        sys.exit(app.exec_())

    confname = options.config_dir
    if confname is None:
        if len(args) == 1:  # for backwards compat, we allow to specify the confname without the "--config-dir" parameter
            confname = args[0]
        else:
            parser.print_help(sys.stderr)
            sys.exit(1)

    if options.fail_proof:
        configRecursionDepth = 0
    else:
        configRecursionDepth = None

    gui = TaurusGui(None, confname=confname,
                    configRecursionDepth=configRecursionDepth)

    gui.show()
    ret = app.exec_()

    taurus.info('Finished execution of TaurusGui')
    sys.exit(ret)


if __name__ == "__main__":
    main()
    # xmlTest()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains the base Object class for taurus."""

__all__ = ["Object"]

__docformat__ = "restructuredtext"


class Object(object):

    def __init__(self):
        pass

    def call__init__(self, klass, *args, **kw):
        """Method to be called from subclasses to call superclass corresponding
        __init__ method. This method ensures that classes from diamond like
        class hierarquies don't call their super classes __init__ more than
        once."""

        if 'inited_class_list' not in self.__dict__:
            self.inited_class_list = []

        if klass not in self.inited_class_list:
            self.inited_class_list.append(klass)
            klass.__init__(self, *args, **kw)

    def call__init__wo_kw(self, klass, *args):
        """Same as call__init__ but without keyword arguments because PyQT does
        not support them."""

        if 'inited_class_list' not in self.__dict__:
            self.inited_class_list = []

        if klass not in self.inited_class_list:
            self.inited_class_list.append(klass)
            klass.__init__(self, *args)

    def getAttrDict(self):
        attr = dict(self.__dict__)
        if 'inited_class_list' in attr:
            del attr['inited_class_list']
        return attr

    def updateAttrDict(self, other):
        attr = other.getAttrDict()
        self.__dict__.update(attr)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
qwtdialog.py: Dialogs for Taurusplot
"""

__all__ = ["TaurusPlotConfigDialog"]

import time

from taurus.external.qt import Qt, Qwt5
from taurus.qt.qtgui.util.ui import UILoadable


# class TaurusPlotConfigCapable:
#    """This class is aimed to act as an interface for class TaurusPlot. Every class that uses
#    TaurusPlotConfigDialog class (as TaurusPlot does) should inherit TaurusPlotConfigCapable class
#    and overwrite the methods defined below in order to avoid firing exceptions"""
#    def __init__(self, parent = None):
#        pass
#
#    def showConfigDialog(self):
#        raise NotImplementedError, "TaurusPlotConfigCapable: method showConfigDialog() not implemented in child class"
#
#    def setAxisScale(self,axis,min,max):
#        raise NotImplementedError, "TaurusPlotConfigCapable: method setAxisScale() not implemented in child class"
#
#    def setAxisAutoScale(self, axis=None):
#        raise NotImplementedError, "TaurusPlotConfigCapable: method setAxisAutoScale() not implemented in child class"
#
#    def setAxisScaleType(self, axis=None, scale=None):
#        raise NotImplementedError, "TaurusPlotConfigCapable: method setAxisScaleType() not implemented in child class"


@UILoadable(with_ui='ui')
class TaurusPlotConfigDialog(Qt.QDialog):
    """This class is used to build and manage the plot configuration dialog. It
    has been designed using the qt designer application, and then loaded to this
    widget. Hence, if you need to modify the dialog, you can use the
    TaurusPlotConfigDialog.ui file (under ui directory) to make it easier."""

    def __init__(self, parent=None, flags=Qt.Qt.WindowFlags()):
        self.parent = parent
        # if parent==None or not isinstance(parent, TaurusPlotConfigCapable):
        #    raise NotImplementedError, "Parent object doesn't implement TaurusPlotConfigCapable class"
        # call qt designer generated functions to initialize GUI
        Qt.QDialog.__init__(self, parent, flags)
        self.loadUi()

        # insert the CurvesAppearanceWidget
        #(@TODO:must be changed to be done directly in the ui, but I couldn't make the widget available to TaurusDesigner)
        from curvesAppearanceChooserDlg import CurvesAppearanceChooser
        layout = Qt.QVBoxLayout()
        self.curvesAppearanceChooser = CurvesAppearanceChooser(None)
        layout.addWidget(self.curvesAppearanceChooser)
        self.ui.curveAppearanceGB.setLayout(layout)

        # set the values for the CurvesAppearanceChooser
        self.curvesAppearanceChooser.setCurves(
            self.parent.getCurveAppearancePropertiesDict())

        # insert valid values into mode combobox (linear or logarihtmic) and
        # set selected one in combobox
        self.ui.xModeComboBox.insertItem(0, "Linear")
        self.ui.xModeComboBox.insertItem(1, "Logarithmic")
        self.ui.y1ModeComboBox.insertItem(0, "Linear")
        self.ui.y1ModeComboBox.insertItem(1, "Logarithmic")
        self.ui.y2ModeComboBox.insertItem(0, "Linear")
        self.ui.y2ModeComboBox.insertItem(1, "Logarithmic")

        # insert valid values into peaks combobox (max, min or both) and set
        # selected one in combobox
        self.ui.peaksComboBox.insertItem(0, "Both")
        self.ui.peaksComboBox.insertItem(1, "Max")
        self.ui.peaksComboBox.insertItem(2, "Min")
        self.ui.peaksComboBox.insertItem(3, "None")

        # Init X axis group
        self.ui.xRangeCB.setVisible(False)
        self.ui.xLabelRange.setVisible(False)
        if self.parent.getXIsTime():
            # adapt the X axis group for time-based measurements
            self.ui.xRangeCB.addItems(["",
                                       "1 m",
                                       "1 h",
                                       "1 d",
                                       "1 w",
                                       "30 d",
                                       "1 y"])
            self.ui.xGroupBox.setTitle("Time")
            self.ui.xLabelMin.setText("Start")
            self.ui.xLabelMax.setText("End")
            timetooltip = """It accepts both absolute data-times (e.g. "25/10/1917 21:45:01") \n"""\
                """or relative ones with format <+|-><number><s|m|h|d|w|y> (e.g. "-12h").\n"""\
                """The keyword "now" is also accepted."""
            self.ui.xEditMin.setToolTip(timetooltip)
            self.ui.xEditMax.setToolTip(timetooltip)
        else:
            # The default settings are ok for non-time plots
            self.ui.xEditMin.setValidator(Qt.QDoubleValidator(self))
            self.ui.xEditMax.setValidator(Qt.QDoubleValidator(self))
            self.ui.xRangeCB.setValidator(Qt.QDoubleValidator(self))
            self.ui.xRangeCB.addItems(["",
                                       "10",
                                       "100",
                                       "1000",
                                       "10000",
                                       "100000",
                                       "1000000"])

        self.ui.xDynScaleCheckBox.setChecked(self.parent.getXDynScale())
        auto = self.parent.axisAutoScale(Qwt5.QwtPlot.xBottom)
        self.ui.xAutoGroupBox.setChecked(not auto)

        # this call already initialises the edit boxes of the X axis
        self.setXDynScale(self.parent.getXDynScale())
        self.ui.xDynScaleCheckBox.setVisible(
            self.parent.isXDynScaleSupported())

        # Init Y axes groups
        self._populateYAxisScales()
        self.ui.y1AutoGroupBox.setChecked(
            not self.parent.axisAutoScale(Qwt5.QwtPlot.yLeft))
        self.ui.y2AutoGroupBox.setChecked(
            not self.parent.axisAutoScale(Qwt5.QwtPlot.yRight))

        # set validators for Y min/max edits
        self.ui.y1EditMin.setValidator(Qt.QDoubleValidator(self))
        self.ui.y1EditMax.setValidator(Qt.QDoubleValidator(self))
        self.ui.y2EditMin.setValidator(Qt.QDoubleValidator(self))
        self.ui.y2EditMax.setValidator(Qt.QDoubleValidator(self))

        # populate initial axis mode (linear or logarithmic) values from plot. Warning!: we're using
        # qwt QwtPlot.axis enum type to relate our GUI combo name and qwt axis
        # name (enum type).
        axes = [self.ui.y1ModeComboBox,
                self.ui.y2ModeComboBox, self.ui.xModeComboBox]
        for axis in range(len(axes)):
            scaleType = self.parent.getAxisTransformationType(axis)
            if scaleType == Qwt5.QwtScaleTransformation.Linear:
                axes[axis].setCurrentIndex(0)
            elif scaleType == Qwt5.QwtScaleTransformation.Log10:
                axes[axis].setCurrentIndex(1)
            else:
                raise TypeError, "TaurusPlotConfigDialog::__init__(): unexpected axis scale type (linear or logarihtmic expected)"
        self.ui.xModeComboBox.setEnabled(not self.parent.getXIsTime())

        # determine which axes are visible
        if not self.parent.axisEnabled(Qwt5.QwtPlot.xBottom):
            self.ui.xGroupBox.setVisible(False)

        # populate initial peaks from parent
        if self.parent._showMaxPeaks and self.parent._showMinPeaks:
            self.ui.peaksComboBox.setCurrentIndex(0)
        elif self.parent._showMaxPeaks:
            self.ui.peaksComboBox.setCurrentIndex(1)
        elif self.parent._showMinPeaks:
            self.ui.peaksComboBox.setCurrentIndex(2)
        else:
            self.ui.peaksComboBox.setCurrentIndex(3)

        # connect signals
        self.ui.buttonBox.button(
            Qt.QDialogButtonBox.Close).clicked.connect(self.hide)
        self.ui.buttonBox.button(
            Qt.QDialogButtonBox.Apply).clicked.connect(self.apply)
        self.ui.xAutoGroupBox.toggled.connect(self.toggledAutoScale)
        self.ui.y1AutoGroupBox.toggled.connect(self.toggledAutoScale)
        self.ui.y2AutoGroupBox.toggled.connect(self.toggledAutoScale)
#        self.connect(self.ui.xEditMin,  Qt.SIGNAL("returnPressed()"),self.apply)
#        self.connect(self.ui.xEditMax,  Qt.SIGNAL("returnPressed()"),self.apply)
#        self.connect(self.ui.y1EditMin, Qt.SIGNAL("returnPressed()"),self.apply)
#        self.connect(self.ui.y1EditMax, Qt.SIGNAL("returnPressed()"),self.apply)
#        self.connect(self.ui.y2EditMin, Qt.SIGNAL("returnPressed()"),self.apply)
#        self.connect(self.ui.y2EditMax, Qt.SIGNAL("returnPressed()"),self.apply)
        self.ui.xModeComboBox.currentIndexChanged.connect(self.modeComboChanged)
        self.ui.xDynScaleCheckBox.toggled.connect(self.setXDynScale)
        #self.connect(self.ui.xRangeCB, Qt.SIGNAL("currentIndexChanged(const QString&)"),self.apply)
        self.ui.y1ModeComboBox.currentIndexChanged.connect(self.modeComboChanged)
        self.ui.y2ModeComboBox.currentIndexChanged.connect(self.modeComboChanged)
        self.ui.peaksComboBox.currentIndexChanged.connect(self.peaksComboChanged)
        # self.connect(self.curvesAppearanceChooser,
        # Qt.SIGNAL("controlChanged"),self.apply) #"autoapply" mode for *all*
        # the curve appearance controls
        self.curvesAppearanceChooser.assignToY1BT.clicked[()].connect(self.setCurvesYAxis)
        self.curvesAppearanceChooser.assignToY2BT.clicked[()].connect(self.setCurvesYAxis)
        self.curvesAppearanceChooser.bckgndBT.clicked.connect(self.changeBackgroundColor)
        self.curvesAppearanceChooser.changeTitlesBT.clicked.connect(self.onChangeTitles)
        self.curvesAppearanceChooser.CurveTitleEdited.connect(self.onCurveTitleEdited)

        # finally adjust size
        self.adjustSize()

    def _populateXAxisScales(self):
        xMin = self.parent.axisScaleDiv(Qwt5.QwtPlot.xBottom).lowerBound()
        xMax = self.parent.axisScaleDiv(Qwt5.QwtPlot.xBottom).upperBound()
        if self.parent.getXIsTime():
            self.ui.xEditMin.setText(time.strftime(
                '%Y/%m/%d %H:%M:%S', time.localtime(int(xMin))))
            self.ui.xEditMax.setText(time.strftime(
                '%Y/%m/%d %H:%M:%S', time.localtime(int(xMax))))
        else:
            self.ui.xEditMin.setText(str(xMin))
            self.ui.xEditMax.setText(str(xMax))

    def _populateYAxisScales(self, axis=None):
        if axis is None or axis == Qwt5.QwtPlot.yLeft:
            self.ui.y1EditMin.setText(
                str(self.parent.axisScaleDiv(Qwt5.QwtPlot.yLeft).lowerBound()))
            self.ui.y1EditMax.setText(
                str(self.parent.axisScaleDiv(Qwt5.QwtPlot.yLeft).upperBound()))
        if axis is None or axis == Qwt5.QwtPlot.yRight:
            self.ui.y2EditMin.setText(
                str(self.parent.axisScaleDiv(Qwt5.QwtPlot.yRight).lowerBound()))
            self.ui.y2EditMax.setText(
                str(self.parent.axisScaleDiv(Qwt5.QwtPlot.yRight).upperBound()))

    def deltatime2str(self, dt, fuzzy=False):
        '''converts a time diff in secs to a string.
        If fuzzy=True it returns an approx time diff in s, min, hours or days'''
        dt = float(dt)
        if not fuzzy:
            return "%g s" % dt
        if dt < 2:
            return "%3.3g s" % dt
        elif dt < 120:
            return "%g s" % round(dt, 0)
        elif dt < 7200:
            return "%g m" % round(dt / 60, 0)
        elif dt < 172800:
            return "%g h" % round(dt / 3600, 0)
        else:
            return "%g d" % round(dt / 86400, 0)

    def str2deltatime(self, strtime):
        '''Translates a time string to seconds
        examples of valid relative times are: "now", "NOW", "Now", "-1d", "3w", "- 3.6e3 s",...
        examples of non-valid relative times:, "now + 2h", "-5", "3H" (unit names are case-sensitive)
        '''
        strtime = str(strtime).strip()
        timeunits = {'s': 1, 'm': 60, 'h': 3600, 'd': 3600 *
                     24, 'w': 3600 * 24 * 7, 'y': 3600 * 24 * 365}
        if strtime.lower() == "now":
            return time.time()
        if strtime[-1] in timeunits.keys():
            try:
                return float(strtime[:-1]) * timeunits[strtime[-1]]
            except Exception, e:
                print '[str2deltatime] No format could be applied to "%s"' % strtime
                return None
        else:
            print '[str2deltatime] Non-supported unit "%s"' % strtime[-1]
            return None

    def strtime2epoch(self, strtime):
        '''
        Translates a str into an epoch value. It accepts "absolute" time notation as well as
        "relative to current time notation" (by expliciting a "+" or "-" prefix)
        (see str2deltatime for relative time notation).

        examples of valid absolute times: "2008-3-25 14:21:59", "25/03/08 14:21", "03-25-2008",...

        It returns None if strtime couldn't be interpreted
        '''
        import time
        t = None
        strtime = str(strtime).strip()

        # try with relative time keywords
        if strtime[0] in ['+', '-']:
            reltime = self.str2deltatime(strtime)
            if reltime is not None:
                return time.time() + reltime

        if strtime.lower() == "now":
            return time.time()

        # try with absolute time
        hour_fts = ['%H:%M:%S', '%H:%M', '']  # ,"%M'",'%S"'
        date_fts = ['%Y-%m-%d', '%y-%m-%d', '%Y/%m/%d', '%y/%m/%d',
                    '%d-%m-%Y', '%d/%m/%Y', '%m-%d-%Y', '%m/%d/%Y']

        tfs = []
        for d in date_fts:
            for h in hour_fts:
                if h:
                    tfs.append('%s %s' % (d, h))
                else:
                    tfs.append(d)

        for tf in tfs:
            try:
                # print 'trying ',tf
                t = time.strptime(strtime, tf)
                if t[0] < 1970:
                    lt = time.localtime()
                    #    (2008,  9,    23,  15,  44,   24,    1,   267, 1)
                    t = (lt[0], lt[1], lt[2], t[3], t[
                         4], lt[5], lt[6], lt[7], lt[8])
                    break
            except Exception, e:
                pass
        if t:
            return time.mktime(t)
        else:
            print 'No format could be applied to "%s"' % strtime
            return None

    def validate(self):
        '''validates the inputs in the dialog.
        If validation is ok, it returns a tuple containing min/max values for each axis (None if they are autoscaled)
        If validation failed, it returns False.

        .. note::
            the values of the max/min boxes are already validated thanks to their attached QDoubleValidators
            (except for xMin/xMax in time Mode, but this case is handled by strtime2epoch)
        '''
        xMin = xMax = y1Min = y1Max = y2Min = y2Max = None

        # Validating X axis values
        xMin = self.parent.axisScaleDiv(Qwt5.QwtPlot.xBottom).lowerBound()
        xMax = self.parent.axisScaleDiv(Qwt5.QwtPlot.xBottom).upperBound()
        if self.ui.xDynScaleCheckBox.isChecked():
            if self.parent.getXIsTime():
                rg = abs(self.str2deltatime(
                    str(self.ui.xRangeCB.currentText())))
                ok1 = (rg is not None)
            else:
                try:
                    rg, ok1 = float(self.ui.xRangeCB.currentText()), True
                except:
                    rg, ok1 = 0.0, False
            xMin = xMax - rg
            if not ok1 or (xMin >= xMax):
                Qt.QMessageBox.warning(
                    self, "Invalid parameters for X axis range")
                return False
        elif self.ui.xAutoGroupBox.isChecked():
            if self.parent.getXIsTime():  # time mode
                xMin = self.strtime2epoch(self.ui.xEditMin.text())
                xMax = self.strtime2epoch(self.ui.xEditMax.text())
                ok1 = (xMin is not None)
                ok2 = (xMax is not None)
            else:  # XY mode
                try:
                    xMin, ok1 = float(self.ui.xEditMin.text()), True
                except:
                    xMin, ok1 = 0.0, False
                try:
                    xMax, ok2 = float(self.ui.xEditMax.text()), True
                except:
                    xMax, ok2 = 0.0, False
            if not (ok1 and ok2) or (xMin >= xMax):
                Qt.QMessageBox.warning(self, "Invalid parameters for X axis",
                                       "Min value must be strictly less than max value")
                return False

        # Validating Y1 axis values
        if self.ui.y1AutoGroupBox.isChecked():
            try:
                y1Min, ok1 = float(self.ui.y1EditMin.text()), True
            except:
                y1Min, ok1 = 0.0, False
            try:
                y1Max, ok2 = float(self.ui.y1EditMax.text()), True
            except:
                y1Max, ok2 = 0.0, False
            if not(ok1 and ok2) or (y1Min >= y1Max):
                Qt.QMessageBox.warning(self, "Invalid parameters for Y1 axis",
                                       "Min value must be strictly less than max value")
                return False

        # Validating Y2 axis values
        if self.ui.y2AutoGroupBox.isChecked():
            try:
                y2Min, ok1 = float(self.ui.y2EditMin.text()), True
            except:
                y2Min, ok1 = 0.0, False
            try:
                y2Max, ok2 = float(self.ui.y2EditMax.text()), True
            except:
                y2Max, ok2 = 0.0, False
            if not (ok1 and ok2) or (y2Min >= y2Max):
                Qt.QMessageBox.warning(self, "Invalid parameters for Y2 axis",
                                       "Min value must be strictly less than max value")
                return False

        # if we reached this point, the input is valid
        return (xMin, xMax, y1Min, y1Max, y2Min, y2Max)

    def apply(self):
        """This will apply the values set in the dialog. Note that some of them
        are not necessary to be set, since they're already are set when changing the item selected"""

        # Apply the changes from the curves Appearance widget (they are already
        # validated).
        prop, names = self.curvesAppearanceChooser.onApply()
        self.parent.onCurveAppearanceChanged(prop, names)

        # validate the rest of the inputs
        v = self.validate()
        if not v:
            return
        xMin, xMax, y1Min, y1Max, y2Min, y2Max = v

        self.parent.setAxisScale(Qwt5.QwtPlot.xBottom, xMin, xMax)
        self.parent.setAxisScale(Qwt5.QwtPlot.yLeft, y1Min, y1Max)
        self.parent.setAxisScale(Qwt5.QwtPlot.yRight, y2Min, y2Max)

        self.parent._xMax, self.parent._xMin = xMax, xMin

    def toggledAutoScale(self, toggled):
        """This will catch the group boxes check/uncheck event, and will enable autoscale in case
        the event has been unchecking 'disable autoscale'"""
        # get and check who fired the signal
        groupBox = self.sender()
        if groupBox is None or not isinstance(groupBox, Qt.QGroupBox):
            self.error(
                "TaurusPlotConfig::toggledAutoScale(): event is not recognized")
            return
        # set the axis autoscale in the corresponding axis
        if groupBox.objectName() == "xAutoGroupBox" and not toggled:
            self.parent.setAxisAutoScale(Qwt5.QwtPlot.xBottom)
        elif groupBox.objectName() == "y1AutoGroupBox" and not toggled:
            self.parent.setAxisAutoScale(Qwt5.QwtPlot.yLeft)
        elif groupBox.objectName() == "y2AutoGroupBox" and not toggled:
            self.parent.setAxisAutoScale(Qwt5.QwtPlot.yRight)

    def modeComboChanged(self, itemSelected):
        """This will catch the combo box selection change and will set the corresponding axis scale to
        the value passed as parameter"""
        # get and check who fired the signal
        comboBox = self.sender()
        if comboBox is None or not isinstance(comboBox, Qt.QComboBox):
            self.error(
                "TaurusPlotConfig::modeComboChanged(): event is not recognized")
            return
        else:
                # set the scale for the corresponding axis
            if comboBox.objectName() == "xModeComboBox":
                if not self.parent.getXIsTime():
                    self.parent.setAxisScaleType(
                        Qwt5.QwtPlot.xBottom, itemSelected)
            elif comboBox.objectName() == "y1ModeComboBox":
                self.parent.setAxisScaleType(Qwt5.QwtPlot.yLeft, itemSelected)
            elif comboBox.objectName() == "y2ModeComboBox":
                self.parent.setAxisScaleType(Qwt5.QwtPlot.yRight, itemSelected)
            else:
                self.error(
                    "TaurusPlotConfig::modeComboChanged() invalid axis: " + comboBox.objectName())

    def setXDynScale(self, checked):
        if checked:
            if self.parent.getXIsTime():
                rg = self.deltatime2str(
                    self.parent.getXAxisRange(), fuzzy=True)
            else:
                rg = "%g" % self.parent.getXAxisRange()
            self.ui.xRangeCB.setEditText(rg)
            # disable autoscale for the X axis
            self.ui.xAutoGroupBox.setChecked(True)
        else:
            self._populateXAxisScales()
        self.parent.setXDynScale(checked)

    def peaksComboChanged(self, itemSelected):
        """This will catch the combo box selection change and will set the corresponding axis to
        show peaks"""
        if itemSelected == 0:
            maxPeak = minPeak = True
        elif itemSelected == 1:
            maxPeak = True
            minPeak = False
        elif itemSelected == 2:
            maxPeak = False
            minPeak = True
        else:
            maxPeak = minPeak = False
        self.parent.showMaxPeaks(maxPeak)
        self.parent.showMinPeaks(minPeak)

    def showCalendar(self, target):
        return
        self.qcalendar = Qt.QCalendarWidget()
        self.qcalendar.show()

    def setCurvesYAxis(self, curvesNamesList=None, axis=None):
        """ calls the parent's setCurvesYAxis method but it automatically determines the parameters if not given"""
        if curvesNamesList is None:
            curvesNamesList = self.curvesAppearanceChooser.getSelectedCurveNames()
        if axis is None:
            sender = self.sender()
            if sender is self.curvesAppearanceChooser.assignToY1BT:
                axis = Qwt5.QwtPlot.yLeft
            elif sender is self.curvesAppearanceChooser.assignToY2BT:
                axis = Qwt5.QwtPlot.yRight
            else:
                raise ValueError("Cannot determine axis from %s" % sender)
        # call the parent's setCurvesYAxis()
        self.parent.setCurvesYAxis(curvesNamesList, axis)
        # update the axis scale edits
        self._populateYAxisScales(axis)

    def changeBackgroundColor(self):
        """Launches a dialog for choosing the parent's canvas background color"""
        plot = self.parent
        color = Qt.QColorDialog.getColor(plot.canvasBackground(), self)
        if Qt.QColor.isValid(color):
            plot.setCanvasBackground(color)
            plot.replot()

    def onChangeTitles(self):
        '''Calls The parent's changeCurvesTitlesDialog method, with the selected
        curves list as the parameter
        '''
        curveNamesList = self.curvesAppearanceChooser.getSelectedCurveNames()
        newTitlesDict = self.parent.changeCurvesTitlesDialog(
            curveNamesList=curveNamesList)
        self.curvesAppearanceChooser.updateTitles(newTitlesDict)

    def onCurveTitleEdited(self, name, newTitle):
        '''slot used when a curve title is edited

        :param name: (str) curve name
        :param name: (QString) new title
        '''
        newTitlesDict = self.parent.setCurvesTitle(newTitle, [name])
        self.curvesAppearanceChooser.updateTitles(newTitlesDict)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Rope patch for better performances. Based on spyderlib.rope_patch"""

__all__ = ["apply"]

__docformat__ = 'restructuredtext'


def apply():
    """Monkey patching rope for better performances"""
    import rope
    if rope.VERSION not in ('0.9.3', '0.9.2'):
        raise ImportError, "rope %s can't be patched" % rope.VERSION

    # Patching pycore.PyCore, so that forced builtin modules (i.e. modules
    # that were declared as 'extension_modules' in rope preferences)
    # will be indeed recognized as builtins by rope, as expected
    from rope.base import pycore

    class PatchedPyCore(pycore.PyCore):

        def get_module(self, name, folder=None):
            """Returns a `PyObject` if the module was found."""
            # check if this is a builtin module
            pymod = self._builtin_module(name)
            if pymod is not None:
                return pymod
            module = self.find_module(name, folder)
            if module is None:
                raise pycore.ModuleNotFoundError(
                    'Module %s not found' % name)
            return self.resource_to_pyobject(module)
    pycore.PyCore = PatchedPyCore

    # Patching BuiltinFunction for the calltip/doc functions to be
    # able to retrieve the function signatures with forced builtins
    from rope.base import builtins, pyobjects
    from spyderlib.utils.dochelpers import getargs

    class PatchedBuiltinFunction(builtins.BuiltinFunction):

        def __init__(self, returned=None, function=None, builtin=None,
                     argnames=[], parent=None):
            builtins._BuiltinElement.__init__(self, builtin, parent)
            pyobjects.AbstractFunction.__init__(self)
            self.argnames = argnames
            if not argnames and builtin:
                self.argnames = getargs(self.builtin)
            if self.argnames is None:
                self.argnames = []
            self.returned = returned
            self.function = function
    builtins.BuiltinFunction = PatchedBuiltinFunction

    # Patching BuiltinName for the go to definition feature to simply work
    # with forced builtins
    from rope.base import libutils
    import inspect

    class PatchedBuiltinName(builtins.BuiltinName):

        def _pycore(self):
            p = self.pyobject
            while p.parent is not None:
                p = p.parent
            if isinstance(p, builtins.BuiltinModule) and p.pycore is not None:
                return p.pycore

        def get_definition_location(self):
            if not inspect.isbuiltin(self.pyobject):
                _lines, lineno = inspect.getsourcelines(self.pyobject.builtin)
                path = inspect.getfile(self.pyobject.builtin)
                pycore = self._pycore()
                if pycore and pycore.project:
                    resource = libutils.path_to_resource(pycore.project, path)
                    module = pyobjects.PyModule(pycore, None, resource)
                    return (module, lineno)
            return (None, None)
    builtins.BuiltinName = PatchedBuiltinName



# -*- coding: utf-8 -*-

##############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
This module provides a simple HTML help browser.
"""

__all__ = ["HelpPanel"]

from taurus.external.qt import Qt
from taurus.external.qt import QtHelp


class _HelpBrowser(Qt.QTextBrowser):

    def __init__(self, help_engine=None, parent=None):
        Qt.QTextBrowser.__init__(self, parent)
        self.__help_engine = None
        if help_engine:
            self.setHelpEngine(help_engine)

    def setHelpEngine(self, help_engine):
        self.__help_engine = help_engine
        content_widget = help_engine.contentWidget()
        index_widget = help_engine.indexWidget()
        content_widget.linkActivated.connect(self.setSource)
        index_widget.linkActivated.connect(self.setSource)

    def loadResource(self, type, url):
        if url.scheme() == "qthelp":
            if self.__help_engine:
                return Qt.QVariant(self.__help_engine.fileData(url))
        return Qt.QTextBrowser.loadResource(self, type, url)


class HelpPanel(Qt.QWidget):
    """
    Simple widget to display application help system. Usage::

        from taurus.external.qt import Qt
        from taurus.qt.qtgui.help import HelpPanel

        app = Qt.QApplication([])
        help_panel = HelpPanel()

        help_panel.setCollectionFile("help_file.qhc")
        help_panel.show()
        app.exec_()
    """

    def __init__(self, collection_file=None, parent=None):
        Qt.QWidget.__init__(self, parent)
        layout = Qt.QVBoxLayout(self)
        self.setLayout(layout)
        self.__help_engine = None
        if collection_file:
            self.setCollectionFile(collection_file)

    def __clear(self):
        layout = self.layout()
        while layout.count():
            layout.takeAt(0)
        self.__help_engine = None

    @Qt.Slot(str)
    def setCollectionFile(self, collection_file):
        """
        Displays the help from the specified collection file

        :param collection_file: the collection file name (.qhc)
        :type collection_file: str
        """

        self.__clear()
        if not collection_file:
            return
        help_engine = QtHelp.QHelpEngine(collection_file, self)
        if not help_engine.setupData():
            raise Exception("Help engine not available")
        layout = self.layout()
        self.__tab = tab = Qt.QTabWidget()
        self.__help_engine = help_engine
        content_widget = help_engine.contentWidget()
        index_widget = help_engine.indexWidget()
        tab.addTab(content_widget, "Contents")
        tab.addTab(index_widget, "Index")
        self.__help_browser = _HelpBrowser(help_engine, self)
        splitter = Qt.QSplitter(Qt.Qt.Horizontal)
        splitter.insertWidget(0, tab)
        splitter.insertWidget(1, self.__help_browser)
        layout.addWidget(splitter)

    def getCollectionFile(self):
        """
        Returns the name of the current collection file or empty
        string if no collection file is active

        :return: the name of the current collection file
        :rtype: str
        """
        if self.__help_engine:
            return self.__help_engine.collectionFile()
        return ""

    def resetCollectionFile(self):
        """
        Resets the collection file
        """
        self.setCollectionFile("")

    #: This property holds the current collection file name
    #:
    #: **Access functions:**
    #:
    #:     * :meth:`HelpPanel.getCollectionFile`
    #:     * :meth:`HelpPanel.setCollectionFile`
    #:     * :meth:`HelpPanel.resetCollectionFile`
    #:
    collectionFile = Qt.Property("QString", getCollectionFile,
                                 setCollectionFile,
                                 resetCollectionFile)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        return {'group': 'Taurus Help',
                'icon': Qt.QIcon.fromTheme("help"),
                'module': 'taurus.qt.qtgui.help',
                'container': False}


def main():
    import sys
    app = Qt.QApplication([])
    help_panel = HelpPanel()
    if len(sys.argv) > 1:
        help_panel.setCollectionFile(sys.argv[1])
    help_panel.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
search.py: methods for getting matching device/attribute/alias names from Tango database

These methods have been borrowed from fandango modules.
"""

import re
import taurus

###############################################################################
# Utils


def searchCl(regexp, target):
    return re.search(extend_regexp(regexp).lower(), target.lower())


def matchCl(regexp, target):
    return re.match(extend_regexp(regexp).lower(), target.lower())


def is_regexp(s):
    return any(c in s for c in '.*[]()+?')


def extend_regexp(s):
    s = str(s).strip()
    if '.*' not in s:
        s = s.replace('*', '.*')
    if '.*' not in s:
        if ' ' in s:
            s = s.replace(' ', '.*')
        if '/' not in s:
            s = '.*' + s + '.*'
    else:
        if not s.startswith('^'):
            s = '^' + s
        if not s.endswith('$'):
            s = s + '$'
    return s


def isString(s):
    typ = s.__class__.__name__.lower()
    return not hasattr(s, '__iter__') and 'str' in typ and 'list' not in typ


def isCallable(obj):
    return hasattr(obj, '__call__')


def isMap(obj):
    return hasattr(obj, 'has_key') or hasattr(obj, 'items')


def isDictionary(obj):
    return isMap(obj)


def isSequence(obj):
    typ = obj.__class__.__name__.lower()
    return (hasattr(obj, '__iter__') or 'list' in typ) and not isString(obj) and not isMap(obj)


def split_model_list(modelNames):
    '''convert str to list if needed (commas and whitespace are considered as separators)'''
    if isString(modelNames):  # isinstance(modelNames,(basestring,Qt.QString)):
        modelNames = str(modelNames).replace(',', ' ')
        modelNames = modelNames.split()
    if isSequence(modelNames):  # isinstance(modelNames,(list.Qt.QStringList)):
        modelNames = [str(s) for s in modelNames]
    return modelNames


def get_matching_devices(expressions, limit=0, exported=False):
    """
    Searches for devices matching expressions, if exported is True only running devices are returned
    """
    db = taurus.Authority()
    all_devs = [s.lower() for s in db.get_device_name('*', '*')]
    # This code is used to get data from multiples hosts
    #if any(not fun.matchCl(rehost,expr) for expr in expressions): all_devs.extend(get_all_devices(exported))
    # for expr in expressions:
    #m = fun.matchCl(rehost,expr)
    # if m:
    #host = m.groups()[0]
    # print 'get_matching_devices(%s): getting %s devices ...'%(expr,host)
    #odb = PyTango.Database(*host.split(':'))
    #all_devs.extend('%s/%s'%(host,d) for d in odb.get_device_name('*','*'))
    result = [e for e in expressions if e.lower() in all_devs]
    expressions = [extend_regexp(e) for e in expressions if e not in result]
    result.extend(filter(lambda d: any(matchCl(extend_regexp(e), d)
                                       for e in expressions), all_devs))
    return result


def get_device_for_alias(alias):
    db = taurus.Authority()
    try:
        return db.get_device_alias(alias)
    except Exception, e:
        if 'no device found' in str(e).lower():
            return None
        return None  # raise e


def get_alias_for_device(dev):
    db = taurus.Authority()
    try:
        # .get_database_device().DbGetDeviceAlias(dev)
        result = db.get_alias(dev)
        return result
    except Exception, e:
        if 'no alias found' in str(e).lower():
            return None
        return None  # raise e


def get_alias_dict(exp='*'):
    tango = taurus.Authority()
    return dict((k, tango.get_device_alias(k)) for k in tango.get_device_alias_list(exp))



'''
Created on Mar 4, 2015

@author: aaron
'''

from decimal import Decimal
import math
import operator

import token as tokenlib

#for controlling order of operations
_OP_PRIORITY = {
    '**' : 3,
    '^' : 3,
    'unary' : 2,
    '*' : 1,
    '' : 1, #operator for implicit ops
    '/' : 1,
    '+' : 0,
    '-' : 0
}

_BINARY_OPERATOR_MAP = {
    '**': operator.pow,
    '*': operator.mul,
    '': operator.mul, #operator for implicit ops
    '/': operator.truediv,
    '+': operator.add,
    '-': operator.sub
}

_UNARY_OPERATOR_MAP = {
    '+': lambda x : x,
    '-': lambda x : x * -1
}


class EvalTreeNode(object):
    
    def __init__(self, left, operator=None, right=None):
        '''
        left + operator + right --> binary op
        left + operator --> unary op
        left + right --> implicit op
        left --> single value
        '''
        self.left = left
        self.operator = operator
        self.right = right
        
    def to_string(self):
        #for debugging purposes
        if self.right:
            comps = [self.left.to_string()]
            if self.operator:
                comps.append(self.operator[1])
            comps.append(self.right.to_string())
        elif self.operator:
            comps = [self.operator[1], self.left.to_string()]
        else:
            return self.left[1]
        return '(%s)' % ' '.join(comps)
    
    def evaluate(self, define_op, bin_op=_BINARY_OPERATOR_MAP, un_op=_UNARY_OPERATOR_MAP):
        '''
        define_op is a callable that translates tokens into objects
        bin_op and un_op provide functions for performing binary and unary operations
        '''
        
        if self.right:
            #binary or implicit operator
            op_text = self.operator[1] if self.operator else ''
            if op_text not in bin_op:
                raise Exception('missing binary operator "%s"' % op_text)
            left = self.left.evaluate(define_op, bin_op, un_op)
            return bin_op[op_text](left, self.right.evaluate(define_op, bin_op, un_op))
        elif self.operator:
            #unary operator
            op_text = self.operator[1]
            if op_text not in un_op:
                raise Exception('missing unary operator "%s"' % op_text)
            return un_op[op_text](self.left.evaluate(define_op, bin_op, un_op))
        else:
            #single value
            return define_op(self.left)
        

def build_eval_tree(tokens, op_priority=_OP_PRIORITY, index=0, depth=0, prev_op=None, ):
    '''
    Params:
    Index, depth, and prev_op used recursively, so don't touch.
    Tokens is an iterable of tokens from an expression to be evaluated.
    
    Transform the tokens from an expression into a recursive parse tree, following order of operations.
    Operations can include binary ops (3 + 4), implicit ops (3 kg), or unary ops (-1).
    
    General Strategy:
    1) Get left side of operator
    2) If no tokens left, return final result
    3) Get operator
    4) Use recursion to create tree starting at token on right side of operator (start at step #1)
    4.1) If recursive call encounters an operator with lower or equal priority to step #2, exit recursion
    5) Combine left side, operator, and right side into a new left side
    6) Go back to step #2
    '''
    if depth == 0 and prev_op == None:
        #ensure tokens is list so we can access by index
        tokens = list(tokens)
        
    result = None
    
    while True:
        current_token = tokens[index]
        token_type = current_token[0]
        token_text = current_token[1]
        
        if token_type == tokenlib.OP:
            if token_text == ')':
                if prev_op == None:
                    raise Exception('unopened parentheses in tokens: %s' % current_token)
                elif prev_op == '(':
                    #close parenthetical group
                    return result, index
                else:
                    #parenthetical group ending, but we need to close sub-operations within group
                    return result, index - 1
            elif token_text == '(':
                #gather parenthetical group
                right, index = build_eval_tree(tokens, op_priority, index+1, 0, token_text)
                if not tokens[index][1] == ')':
                    raise Exception('weird exit from parentheses')
                if result:
                    #implicit op with a parenthetical group, i.e. "3 (kg ** 2)"
                    result = EvalTreeNode(left=result, right=right)
                else:
                    #get first token
                    result = right
            elif token_text in op_priority:
                if result:
                    #equal-priority operators are grouped in a left-to-right order, unless they're
                    #exponentiation, in which case they're grouped right-to-left
                    #this allows us to get the expected behavior for multiple exponents 
                    #    (2^3^4)  --> (2^(3^4))
                    #    (2 * 3 / 4) --> ((2 * 3) / 4)
                    if op_priority[token_text] <= op_priority.get(prev_op, -1) and token_text not in ['**', '^']:
                        #previous operator is higher priority, so end previous binary op
                        return result, index - 1
                    #get right side of binary op
                    right, index = build_eval_tree(tokens, op_priority, index+1, depth+1, token_text)
                    result = EvalTreeNode(left=result, operator=current_token, right=right)
                else:
                    #unary operator
                    right, index = build_eval_tree(tokens, op_priority, index+1, depth+1, 'unary')
                    result = EvalTreeNode(left=right, operator=current_token)
        elif token_type == tokenlib.NUMBER or token_type == tokenlib.NAME:
            if result:
                #tokens with an implicit operation i.e. "1 kg"
                if op_priority[''] <= op_priority.get(prev_op, -1):
                    #previous operator is higher priority than implicit, so end previous binary op
                    return result, index - 1
                right, index = build_eval_tree(tokens, op_priority, index, depth+1, '')
                result = EvalTreeNode(left=result, right=right)
            else:
                #get first token
                result = EvalTreeNode(left=current_token)
        
        if tokens[index][0] == tokenlib.ENDMARKER:
            if prev_op == '(':
                raise Exception('unclosed parentheses in tokens')
            if depth > 0 or prev_op:
                #have to close recursion
                return result, index
            else:
                #recursion all closed, so just return the final result
                return result
            
        if index + 1 >= len(tokens):
            #should hit ENDMARKER before this ever happens
            raise Exception('unexpected end to tokens')
        
        index += 1

def _test_build_tree(input_text):
    '''
    
    ####
    >>> _test_build_tree('3') #single number
    u'3'
    >>> _test_build_tree('1 + 2') #basic addition
    u'(1 + 2)'
    >>> _test_build_tree('2 * 3 + 4') #order of operations
    u'((2 * 3) + 4)'
    >>> _test_build_tree('2 * (3 + 4)') #parentheses
    u'(2 * (3 + 4))'
    >>> _test_build_tree('1 + 2 * 3 ** (4 + 3 / 5)') #more order of operations
    u'(1 + (2 * (3 ** (4 + (3 / 5)))))'
    >>> _test_build_tree('1 * ((3 + 4) * 5)') #nested parentheses at beginning
    u'(1 * ((3 + 4) * 5))'
    >>> _test_build_tree('1 * (5 * (3 + 4))') #nested parentheses at end
    u'(1 * (5 * (3 + 4)))'
    >>> _test_build_tree('1 * (5 * (3 + 4) / 6)') #nested parentheses in middle
    u'(1 * ((5 * (3 + 4)) / 6))'
    >>> _test_build_tree('-1') #unary
    u'(- 1)'
    >>> _test_build_tree('3 * -1') #unary
    u'(3 * (- 1))'
    >>> _test_build_tree('3 * --1') #double unary
    u'(3 * (- (- 1)))'
    >>> _test_build_tree('3 * -(2 + 4)') #parenthetical unary
    u'(3 * (- (2 + 4)))'
    >>> _test_build_tree('3 * -((2 + 4))') #parenthetical unary
    u'(3 * (- (2 + 4)))'
    >>> _test_build_tree('3 4') #implicit op
    u'(3 4)'
    >>> _test_build_tree('3 (2 + 4)') #implicit op, then parentheses
    u'(3 (2 + 4))'
    >>> _test_build_tree('(3 ** 4 ) 5') #parentheses, then implicit
    u'((3 ** 4) 5)'
    >>> _test_build_tree('3 4 ** 5') #implicit op, then exponentiation
    u'(3 (4 ** 5))'
    >>> _test_build_tree('3 4 + 5') #implicit op, then addition
    u'((3 4) + 5)'
    >>> _test_build_tree('3 ** 4 5') #power followed by implicit
    u'((3 ** 4) 5)'
    >>> _test_build_tree('3 (4 ** 5)') #implicit with parentheses
    u'(3 (4 ** 5))'
    >>> _test_build_tree('3e-1') #exponent with e
    u'3e-1'
    
    >>> _test_build_tree('kg ** 1 * s ** 2') #multiple units with exponents
    u'((kg ** 1) * (s ** 2))'
    >>> _test_build_tree('kg ** -1 * s ** -2') #multiple units with neg exponents
    u'((kg ** (- 1)) * (s ** (- 2)))'
    >>> _test_build_tree('kg^-1 * s^-2') #multiple units with neg exponents
    u'((kg ^ (- 1)) * (s ^ (- 2)))'
    >>> _test_build_tree('kg^-1 s^-2') #multiple units with neg exponents, implicit op
    u'((kg ^ (- 1)) (s ^ (- 2)))'
    
    >>> _test_build_tree('2 ^ 3 ^ 2') #nested power
    u'(2 ^ (3 ^ 2))'
    
    >>> _test_build_tree('gram * second / meter ** 2') #nested power
    u'((gram * second) / (meter ** 2))'
    >>> _test_build_tree('gram / meter ** 2 / second') #nested power
    u'((gram / (meter ** 2)) / second)'
    
    #units should behave like numbers, so we don't need a bunch of extra tests for them
    >>> _test_build_tree('3 kg + 5') #implicit op, then addition
    u'((3 kg) + 5)'
    '''
    return build_eval_tree(tokenizer(input_text)).to_string()

if __name__ == "__main__":
    import doctest, pint
    from pint.compat import tokenizer
    doctest.testmod()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
This module contains some Taurus-wide default configurations.

The idea is that the final user may edit the values here to customize certain
aspects of Taurus.
"""

# A map for using custom widgets for certain devices in TaurusForms. It is a
# dictionary with the following structure:
# device_class_name:(classname_with_full_module_path, args, kwargs)
# where the args and kwargs will be passed to the constructor of the class
T_FORM_CUSTOM_WIDGET_MAP = \
    {'SimuMotor': ('sardana.taurus.qt.qtgui.extra_pool.PoolMotorTV', (), {}),
     'Motor': ('sardana.taurus.qt.qtgui.extra_pool.PoolMotorTV', (), {}),
     'PseudoMotor': ('sardana.taurus.qt.qtgui.extra_pool.PoolMotorTV', (), {}),
     'PseudoCounter': ('sardana.taurus.qt.qtgui.extra_pool.PoolChannelTV', (), {}),
     'CTExpChannel': ('sardana.taurus.qt.qtgui.extra_pool.PoolChannelTV', (), {}),
     'ZeroDExpChannel': ('sardana.taurus.qt.qtgui.extra_pool.PoolChannelTV', (), {}),
     'OneDExpChannel': ('sardana.taurus.qt.qtgui.extra_pool.PoolChannelTV', (), {}),
     'TwoDExpChannel': ('sardana.taurus.qt.qtgui.extra_pool.PoolChannelTV', (), {}),
     'IORegister': ('sardana.taurus.qt.qtgui.extra_pool.PoolIORegisterTV', (), {})
     }

# Compact mode for widgets
# True sets the preferred mode of TaurusForms to use "compact" widgets
T_FORM_COMPACT = False

# Strict RFC3986 URI names in models
# True makes Taurus only use the strict URI names
# False enables a backwards-compatibility mode for pre-sep3 model names
STRICT_MODEL_NAMES = False


# Lightweight imports:
# True enables delayed imports (may break older code).
# False (or commented out) for backwards compatibility
LIGHTWEIGHT_IMPORTS = False

# Set your default scheme (if not defined, "tango" is assumed)
DEFAULT_SCHEME = "tango"

# Extra Taurus schemes. You can add a list of modules to be loaded for
# providing support to new schemes
# EXTRA_SCHEME_MODULES = ['myownschememodule']

# ----------------------------------------------------------------------------
# PLY (lex/yacc) optimization: 1=Active (default) , 0=disabled.
# Set PLY_OPTIMIZE = 0 if you are getting yacc exceptions while loading
# synoptics
# ----------------------------------------------------------------------------

PLY_OPTIMIZE = 1

# ----------------------------------------------------------------------------
# Taurus namespace
# ----------------------------------------------------------------------------

NAMESPACE = 'taurus'

# ----------------------------------------------------------------------------
# Qt configuration
# ----------------------------------------------------------------------------

#: Auto initialize Qt
DEFAULT_QT_AUTO_INIT = True

#: Set preffered API if not is already loaded
DEFAULT_QT_AUTO_API = 'PyQt4'

#: Whether or not should be strict in choosing Qt API
DEFAULT_QT_AUTO_STRICT = False

#: Auto initialize Qt logging to python logging
DEFAULT_QT_AUTO_INIT_LOG = True

#: Auto initialize taurus resources (icons)
DEFAULT_QT_AUTO_INIT_RES = True

#: Remove input hook (only valid for PyQt4)
DEFAULT_QT_AUTO_REMOVE_INPUTHOOK = True

#: Auto initialize Qt
QT_AUTO_INIT = DEFAULT_QT_AUTO_INIT

#: Set preffered API if not is already loaded
QT_AUTO_API = DEFAULT_QT_AUTO_API

#: Whether or not should be strict in choosing Qt API
QT_AUTO_STRICT = DEFAULT_QT_AUTO_STRICT

#: Auto initialize Qt logging to python logging
QT_AUTO_INIT_LOG = DEFAULT_QT_AUTO_INIT_LOG

#: Auto initialize taurus resources (icons)
QT_AUTO_INIT_RES = DEFAULT_QT_AUTO_INIT_RES

#: Remove input hook (only valid for PyQt4)
QT_AUTO_REMOVE_INPUTHOOK = DEFAULT_QT_AUTO_REMOVE_INPUTHOOK

#: Select the theme to be used: set the theme dir  and the theme name.
#: The path can be absolute or relative to the dir of taurus.qt.qtgui.icon
#: If not set, the dir of taurus.qt.qtgui.icon will be used
QT_THEME_DIR = ''
#: The name of the icon theme (e.g. 'Tango', 'Oxygen', etc). Default='Tango'
QT_THEME_NAME = 'Tango'
#: In Linux the QT_THEME_NAME is not applied (to respect the system theme)
#: setting QT_THEME_FORCE_ON_LINUX=True overrides this.
QT_THEME_FORCE_ON_LINUX = True


# ----------------------------------------------------------------------------
# Deprecation handling:
# Note: this API is still experimental and may be subject to change
# (hence the "_" in the options)
# ----------------------------------------------------------------------------

# set the maximum number of same-message deprecations to be logged.
# None (or not set) indicates no limit. -1 indicates that an exception should
# be raised instead of logging the message (useful for finding obsolete code)
_MAX_DEPRECATIONS_LOGGED = 1

# Custom organization logo. Set the absolute path to an image file to be used as your
# organization logo. Qt registered paths can also be used. 
# If not set, it defaults to 'logos:taurus.png" 
# (note that "logos:" is a Qt a registered path for "<taurus>/qt/qtgui/icon/logos/")
# ORGANIZATION_LOGO = "logos:taurus.png"
